import numpy as np
# version: Apr 22, 2025

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.6"

class SingleSFFTConfigure_Cupy:
    @staticmethod
    def SSCC(NX, NY, KerHW, KerPolyOrder=2, BGPolyOrder=2, ConstPhotRatio=True, VERBOSE_LEVEL=2):
        
        import cupy as cp
        N0, N1 = int(NX), int(NY)
        w0, w1 = int(KerHW), int(KerHW)
        DK, DB = int(KerPolyOrder), int(BGPolyOrder)
        MaxThreadPerB = 8     # FIXME emperical value
        SFFTParam_dict = {}
        SFFTModule_dict = {}
        
        if DK not in [0, 1, 2, 3]:
            # NOTE DK = 0 means a spatial invariant matching-kernel.
            raise Exception('MeLOn ERROR: Input KerPolyOrder should be 0/1/2/3!')

        if DB not in [0, 1, 2, 3]:
            # NOTE DB = 0 means a flat differential-background, however, we cannot kill the sky term in our script.
            raise Exception('MeLOn ERROR: Input BGPolyOrder should be 0/1/2/3!')
        
        if (N0 < MaxThreadPerB) or (N1 < MaxThreadPerB):
            raise Exception('MeLOn ERROR: Input Image has dramatically small size!')

        if VERBOSE_LEVEL in [1, 2]:
            print('\n --//--//--//--//-- TRIGGER SFFT COMPILATION --//--//--//--//-- ')
            print('\n ---//--- KerPolyOrder %d | BGPolyOrder %d | KerHW [%d] ---//--- ' %(DK, DB, w0))

        # * Make a dictionary for SFFT parameters
        L0 = 2*w0+1                       # matching-kernel XSize
        L1 = 2*w1+1                       # matching-kernel YSize
        Fab = L0 * L1                     # dof for index 𝛼𝛽
        Fij = int((DK+1)*(DK+2)/2)        # dof for matching-kernel polynomial index ij 
        Fpq = int((DB+1)*(DB+2)/2)        # dof for differential-background polynomial index pq 
        SCALE = np.float64(1/(N0*N1))     # Scale of Image-Size
        SCALE_L = np.float64(1/SCALE)     # Reciprocal Scale of Image Size

        NEQ = Fij*Fab+Fpq                         # Linear-System side-length
        Fijab = Fij * Fab                         # Linear-System Major side-length
        NEQ_FSfree = NEQ - (Fij - 1)              # Forbidden-Stripes-Free Linear-System side-length

        FOMG, FGAM, FTHE = Fij**2, Fij*Fpq, Fij   # OMG / GAM / THE has shape (dof, N0, N1)
        FPSI, FPHI, FDEL = Fpq*Fij, Fpq**2, Fpq   # PSI / PHI / DEL has shape (dof, N0, N1)

        SFFTParam_dict['N0'] = N0
        SFFTParam_dict['N1'] = N1
        SFFTParam_dict['w0'] = w0
        SFFTParam_dict['w1'] = w1
        SFFTParam_dict['DK'] = DK
        SFFTParam_dict['DB'] = DB
        SFFTParam_dict['ConstPhotRatio'] = ConstPhotRatio
        SFFTParam_dict['MaxThreadPerB'] = MaxThreadPerB

        SFFTParam_dict['L0'] = L0
        SFFTParam_dict['L1'] = L1
        SFFTParam_dict['Fab'] = Fab
        SFFTParam_dict['Fij'] = Fij
        SFFTParam_dict['Fpq'] = Fpq
        SFFTParam_dict['SCALE'] = SCALE
        SFFTParam_dict['SCALE_L'] = SCALE_L
        SFFTParam_dict['NEQ'] = NEQ
        SFFTParam_dict['Fijab'] = Fijab
        SFFTParam_dict['NEQ_FSfree'] = NEQ_FSfree

        SFFTParam_dict['FOMG'] = FOMG
        SFFTParam_dict['FGAM'] = FGAM
        SFFTParam_dict['FTHE'] = FTHE
        SFFTParam_dict['FPSI'] = FPSI
        SFFTParam_dict['FPHI'] = FPHI
        SFFTParam_dict['FDEL'] = FDEL

        # * Load SFFT CUDA modules
        #   NOTE: Generally, a kernel function is defined without knowledge about Grid-Block-Thread Management.
        #         However, we need to know the size of threads per block if SharedMemory is called.

        # ************************************ SpatialCoor.cu ************************************ #

        # ** produce spatial coordinate X/Y/oX/oY-map
        _refdict = {'N0': N0, 'N1': N1}
        _funcstr = r"""
        extern "C" __global__ void kmain(int PixA_X_GPU[%(N0)s][%(N1)s], int PixA_Y_GPU[%(N0)s][%(N1)s], 
            double PixA_CX_GPU[%(N0)s][%(N1)s], double PixA_CY_GPU[%(N0)s][%(N1)s])
        {
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;
            
            int N0 = %(N0)s;
            int N1 = %(N1)s;
            
            if (ROW < N0 && COL < N1) {
                double cx = (ROW + 1.0) / N0;  
                double cy = (COL + 1.0) / N1;
                PixA_X_GPU[ROW][COL] = ROW;
                PixA_Y_GPU[ROW][COL] = COL;
                PixA_CX_GPU[ROW][COL] = cx;
                PixA_CY_GPU[ROW][COL] = cy;
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=False)
        SFFTModule_dict['SpatialCoor'] = _module

        # ************************************ SpatialPoly.cu ************************************ #

        # ** produce Iij, Tpq
        _refdict = {'N0': N0, 'N1': N1, 'Fij': Fij, 'Fpq': Fpq}
        _funcstr = r"""
        extern "C" __global__ void kmain(int REF_ij_GPU[%(Fij)s][2], int REF_pq_GPU[%(Fpq)s][2],
            double PixA_CX_GPU[%(N0)s][%(N1)s], double PixA_CY_GPU[%(N0)s][%(N1)s], double PixA_I_GPU[%(N0)s][%(N1)s], 
            double SPixA_Iij_GPU[%(Fij)s][%(N0)s][%(N1)s], double SPixA_Tpq_GPU[%(Fpq)s][%(N0)s][%(N1)s]) 
        {
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;
        
            int N0 = %(N0)s;
            int N1 = %(N1)s;
            int Fij = %(Fij)s;
            int Fpq = %(Fpq)s;
            
            if (ROW < N0 && COL < N1) {
                for(int ij = 0; ij < Fij; ++ij) {
                    int i = REF_ij_GPU[ij][0];
                    int j = REF_ij_GPU[ij][1];
                    double poly_kterm = pow(PixA_CX_GPU[ROW][COL], i) * pow(PixA_CY_GPU[ROW][COL], j);
                    SPixA_Iij_GPU[ij][ROW][COL] = PixA_I_GPU[ROW][COL] * poly_kterm;
                }
                    
                for(int pq = 0; pq < Fpq; ++pq) {
                    int p = REF_pq_GPU[pq][0];
                    int q = REF_pq_GPU[pq][1];
                    double poly_bterm = pow(PixA_CX_GPU[ROW][COL], p) * pow(PixA_CY_GPU[ROW][COL], q);
                    SPixA_Tpq_GPU[pq][ROW][COL] = poly_bterm;
                }
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=False)
        SFFTModule_dict['SpatialPoly'] = _module

        # ************************************ HadProd_OMG.cu & FillLS_OMG.cu ************************************ #

        # ** Hadamard Product [𝛀]
        _refdict = {'N0': N0, 'N1': N1, 'Fij': Fij, 'FOMG': FOMG, 'ThreadPerB': MaxThreadPerB}
        _funcstr = r"""
        #include "cuComplex.h"
        extern "C" __global__ void kmain(int SREF_iji0j0_GPU[%(FOMG)s][2],
            cuDoubleComplex SPixA_FIij_GPU[%(Fij)s][%(N0)s][%(N1)s],
            cuDoubleComplex SPixA_CFIij_GPU[%(Fij)s][%(N0)s][%(N1)s], 
            cuDoubleComplex HpOMG_GPU[%(FOMG)s][%(N0)s][%(N1)s]) 
        {
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;

            int N0 = %(N0)s;
            int N1 = %(N1)s;
            int Fij = %(Fij)s;
            int FOMG = %(FOMG)s;

            __shared__ cuDoubleComplex ShSPixA_FIij_GPU[%(Fij)s][%(ThreadPerB)s][%(ThreadPerB)s];
            if (ROW < N0 && COL < N1) {
                for(int ij = 0; ij < Fij; ++ij){
                    ShSPixA_FIij_GPU[ij][threadIdx.x][threadIdx.y] = SPixA_FIij_GPU[ij][ROW][COL];
                }
            }

            __shared__ cuDoubleComplex ShSPixA_CFIij_GPU[%(Fij)s][%(ThreadPerB)s][%(ThreadPerB)s];
            if (ROW < N0 && COL < N1) {
                for(int ij = 0; ij < Fij; ++ij){
                    ShSPixA_CFIij_GPU[ij][threadIdx.x][threadIdx.y] = SPixA_CFIij_GPU[ij][ROW][COL];
                }
            }
            __syncthreads();

            if (ROW < N0 && COL < N1) {
                for(int i8j8ij = 0; i8j8ij < FOMG; ++i8j8ij){
                    
                    int i8j8 = SREF_iji0j0_GPU[i8j8ij][0];
                    int ij = SREF_iji0j0_GPU[i8j8ij][1];

                    HpOMG_GPU[i8j8ij][ROW][COL] = cuCmul(ShSPixA_FIij_GPU[i8j8][threadIdx.x][threadIdx.y], 
                        ShSPixA_CFIij_GPU[ij][threadIdx.x][threadIdx.y]);
                }
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=True)
        SFFTModule_dict['HadProd_OMG'] = _module

        # ** Fill Linear-System [𝛀]
        _refdict = {'N0': N0, 'N1': N1, 'Fijab': Fijab, 'Fab': Fab, 'Fij': Fij, 'NEQ': NEQ, 'FOMG': FOMG}
        _funcstr = r"""
        extern "C" __global__ void kmain(int SREF_ijab_GPU[%(Fijab)s][2], int REF_ab_GPU[%(Fab)s][2], 
            double PreOMG_GPU[%(FOMG)s][%(N0)s][%(N1)s], double LHMAT_GPU[%(NEQ)s][%(NEQ)s]) 
        {
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;
            
            int N0 = %(N0)s;
            int N1 = %(N1)s;
            int Fij = %(Fij)s;
            int Fijab = %(Fijab)s;
            
            if (ROW < Fijab && COL < Fijab) {
                
                // INDEX Analysis
                int i8j8 = SREF_ijab_GPU[ROW][0];
                int a8b8 = SREF_ijab_GPU[ROW][1];
                int ij = SREF_ijab_GPU[COL][0];
                int ab = SREF_ijab_GPU[COL][1];

                int a8 = REF_ab_GPU[a8b8][0];
                int b8 = REF_ab_GPU[a8b8][1];
                int a = REF_ab_GPU[ab][0];
                int b = REF_ab_GPU[ab][1];
                int idx = i8j8 * Fij + ij;

                // Define Mod_N0(rho), Mod_N1(eps)
                float tmp = 0.0;
                tmp = fmod(float(a8), float(N0));
                if (tmp < 0.0) {tmp += float(N0);}
                int MODa8 = tmp;

                tmp = fmod(float(b8), float(N1));
                if (tmp < 0.0) {tmp += float(N1);}
                int MODb8 = tmp;

                tmp = fmod(float(-a), float(N0));
                if (tmp < 0.0) {tmp += float(N0);}
                int MOD_a = tmp;

                tmp = fmod(float(-b), float(N1));
                if (tmp < 0.0) {tmp += float(N1);}
                int MOD_b = tmp;

                tmp = fmod(float(a8-a), float(N0));
                if (tmp < 0.0) {tmp += float(N0);}
                int MODa8_a = tmp;

                tmp = fmod(float(b8-b), float(N1));
                if (tmp < 0.0) {tmp += float(N1);}
                int MODb8_b = tmp;

                // Fill Linear System [A-component]
                if ((a8 != 0 || b8 != 0) && (a != 0 || b != 0)) {
                    LHMAT_GPU[ROW][COL] = - PreOMG_GPU[idx][MODa8][MODb8]
                                          - PreOMG_GPU[idx][MOD_a][MOD_b] 
                                          + PreOMG_GPU[idx][MODa8_a][MODb8_b] 
                                          + PreOMG_GPU[idx][0][0];
                }

                if ((a8 == 0 && b8 == 0) && (a != 0 || b != 0)) {
                    LHMAT_GPU[ROW][COL] = PreOMG_GPU[idx][MOD_a][MOD_b] - PreOMG_GPU[idx][0][0];
                }

                if ((a8 != 0 || b8 != 0) && (a == 0 && b == 0)) {
                    LHMAT_GPU[ROW][COL] = PreOMG_GPU[idx][MODa8][MODb8] - PreOMG_GPU[idx][0][0];
                }

                if ((a8 == 0 && b8 == 0) && (a == 0 && b == 0)) {
                    LHMAT_GPU[ROW][COL] = PreOMG_GPU[idx][0][0];
                }
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=False)
        SFFTModule_dict['FillLS_OMG'] = _module

        # ************************************ HadProd_GAM.cu & FillLS_GAM.cu ************************************ #

        # ** Hadamard Product [𝜦]
        _refdict = {'N0': N0, 'N1': N1, 'Fij': Fij, 'Fpq': Fpq, 'FGAM': FGAM, 'ThreadPerB': MaxThreadPerB}
        _funcstr = r"""
        #include "cuComplex.h"
        extern "C" __global__ void kmain(int SREF_ijpq_GPU[%(FGAM)s][2], 
            cuDoubleComplex SPixA_FIij_GPU[%(Fij)s][%(N0)s][%(N1)s],
            cuDoubleComplex SPixA_CFTpq_GPU[%(Fpq)s][%(N0)s][%(N1)s], 
            cuDoubleComplex HpGAM_GPU[%(FGAM)s][%(N0)s][%(N1)s])
        {
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;

            int N0 = %(N0)s;
            int N1 = %(N1)s;
            int Fij = %(Fij)s;
            int Fpq = %(Fpq)s;
            int FGAM = %(FGAM)s;
            
            __shared__ cuDoubleComplex ShSPixA_FIij_GPU[%(Fij)s][%(ThreadPerB)s][%(ThreadPerB)s];
            if (ROW < N0 && COL < N1) {
                for(int ij = 0; ij < Fij; ++ij){
                    ShSPixA_FIij_GPU[ij][threadIdx.x][threadIdx.y] = SPixA_FIij_GPU[ij][ROW][COL];
                }
            }

            __shared__ cuDoubleComplex ShSPixA_CFTpq_GPU[%(Fpq)s][%(ThreadPerB)s][%(ThreadPerB)s];
            if (ROW < N0 && COL < N1) {
                for(int pq = 0; pq < Fpq; ++pq){
                    ShSPixA_CFTpq_GPU[pq][threadIdx.x][threadIdx.y] = SPixA_CFTpq_GPU[pq][ROW][COL];
                }
            }
            __syncthreads();
            
            if (ROW < N0 && COL < N1) {        
                for(int i8j8pq = 0; i8j8pq < FGAM; ++i8j8pq){

                    int i8j8 = SREF_ijpq_GPU[i8j8pq][0];
                    int pq = SREF_ijpq_GPU[i8j8pq][1];

                    HpGAM_GPU[i8j8pq][ROW][COL] = cuCmul(ShSPixA_FIij_GPU[i8j8][threadIdx.x][threadIdx.y], 
                        ShSPixA_CFTpq_GPU[pq][threadIdx.x][threadIdx.y]);
                }
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=True)
        SFFTModule_dict['HadProd_GAM'] = _module

        # ** Fill Linear-System [𝜦]
        _refdict = {'N0': N0, 'N1': N1, 'NEQ': NEQ, 'Fab': Fab, 'Fpq': Fpq, 'Fijab': Fijab, 'FGAM': FGAM}
        _funcstr = r"""
        extern "C" __global__ void kmain(int SREF_ijab_GPU[%(Fijab)s][2], int REF_ab_GPU[%(Fab)s][2], 
            double PreGAM_GPU[%(FGAM)s][%(N0)s][%(N1)s], double LHMAT_GPU[%(NEQ)s][%(NEQ)s]) 
        {
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;

            int N0 = %(N0)s;
            int N1 = %(N1)s;
            int Fpq = %(Fpq)s;
            int Fijab = %(Fijab)s;
            
            if (ROW < Fijab && COL < Fpq) {
                
                // INDEX Analysis
                int i8j8 = SREF_ijab_GPU[ROW][0];
                int a8b8 = SREF_ijab_GPU[ROW][1];
                int pq = COL;

                int a8 = REF_ab_GPU[a8b8][0];
                int b8 = REF_ab_GPU[a8b8][1];
                int idx = i8j8 * Fpq + pq;
                int cCOL = Fijab + COL;              // add offset

                // Define Mod_N0(rho), Mod_N1(eps)
                float tmp = 0.0;
                tmp = fmod(float(a8), float(N0));
                if (tmp < 0.0) {tmp += float(N0);}
                int MODa8 = tmp;

                tmp = fmod(float(b8), float(N1));
                if (tmp < 0.0) {tmp += float(N1);}
                int MODb8 = tmp;

                // Fill Linear System [B-component]
                if (a8 != 0 || b8 != 0) {
                    LHMAT_GPU[ROW][cCOL] = PreGAM_GPU[idx][MODa8][MODb8] - PreGAM_GPU[idx][0][0];
                }

                if (a8 == 0 && b8 == 0) {
                    LHMAT_GPU[ROW][cCOL] = PreGAM_GPU[idx][0][0];
                }
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=False)
        SFFTModule_dict['FillLS_GAM'] = _module

        # ************************************ HadProd_PSI.cu & FillLS_PSI.cu ************************************ #

        # ** Hadamard Product [𝜳]
        _refdict = {'N0': N0, 'N1': N1, 'Fij': Fij, 'Fpq': Fpq, 'FPSI': FPSI, 'ThreadPerB': MaxThreadPerB}
        _funcstr = r"""
        #include "cuComplex.h"
        extern "C" __global__ void kmain(int SREF_pqij_GPU[%(FPSI)s][2],
            cuDoubleComplex SPixA_CFIij_GPU[%(Fij)s][%(N0)s][%(N1)s], 
            cuDoubleComplex SPixA_FTpq_GPU[%(Fpq)s][%(N0)s][%(N1)s], 
            cuDoubleComplex HpPSI_GPU[%(FPSI)s][%(N0)s][%(N1)s]) 
        {    
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;

            int N0 = %(N0)s;
            int N1 = %(N1)s;
            int Fpq = %(Fpq)s;
            int Fij = %(Fij)s;
            int FPSI = %(FPSI)s;
            
            __shared__ cuDoubleComplex ShSPixA_FTpq_GPU[%(Fpq)s][%(ThreadPerB)s][%(ThreadPerB)s];
            if (ROW < N0 && COL < N1) {
                for(int pq = 0; pq < Fpq; ++pq){
                    ShSPixA_FTpq_GPU[pq][threadIdx.x][threadIdx.y] = SPixA_FTpq_GPU[pq][ROW][COL];
                }
            }

            __shared__ cuDoubleComplex ShSPixA_CFIij_GPU[%(Fij)s][%(ThreadPerB)s][%(ThreadPerB)s];
            if (ROW < N0 && COL < N1) {
                for(int ij = 0; ij < Fij; ++ij){
                    ShSPixA_CFIij_GPU[ij][threadIdx.x][threadIdx.y] = SPixA_CFIij_GPU[ij][ROW][COL];
                }
            }
            __syncthreads();

            if (ROW < N0 && COL < N1) {
                for(int p8q8ij = 0; p8q8ij < FPSI; ++p8q8ij){
                    
                    int p8q8 = SREF_pqij_GPU[p8q8ij][0];
                    int ij = SREF_pqij_GPU[p8q8ij][1];

                    HpPSI_GPU[p8q8ij][ROW][COL] = cuCmul(ShSPixA_FTpq_GPU[p8q8][threadIdx.x][threadIdx.y], 
                        ShSPixA_CFIij_GPU[ij][threadIdx.x][threadIdx.y]);
                }
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=True)
        SFFTModule_dict['HadProd_PSI'] = _module

        # ** Fill Linear-System [𝜳]
        _refdict = {'N0': N0, 'N1': N1, 'NEQ': NEQ, 'Fab': Fab, 'Fij': Fij, 'Fpq': Fpq, 'Fijab': Fijab, 'FPSI': FPSI}
        _funcstr = r"""
        extern "C" __global__ void kmain(int SREF_ijab_GPU[%(Fijab)s][2], int REF_ab_GPU[%(Fab)s][2], 
            double PrePSI_GPU[%(FPSI)s][%(N0)s][%(N1)s], double LHMAT_GPU[%(NEQ)s][%(NEQ)s]) 
        {
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;
            
            int N0 = %(N0)s;
            int N1 = %(N1)s;
            int Fij = %(Fij)s;
            int Fpq = %(Fpq)s;
            int Fijab = %(Fijab)s;

            if (ROW < Fpq && COL < Fijab) {

                // INDEX Analysis
                int cROW = Fijab + ROW;              // add offset
                int p8q8 = ROW;
                int ij = SREF_ijab_GPU[COL][0];
                int ab = SREF_ijab_GPU[COL][1];
                int a = REF_ab_GPU[ab][0];
                int b = REF_ab_GPU[ab][1];
                int idx = p8q8 * Fij + ij;

                // Define Mod_N0(rho), Mod_N1(eps)
                float tmp = 0.0;
                tmp = fmod(float(-a), float(N0));
                if (tmp < 0.0) {tmp += float(N0);}
                int MOD_a = tmp;

                tmp = fmod(float(-b), float(N1));
                if (tmp < 0.0) {tmp += float(N1);}
                int MOD_b = tmp;
                
                // Fill Linear System [B#-component]
                if (a != 0 || b != 0) {
                    LHMAT_GPU[cROW][COL] = PrePSI_GPU[idx][MOD_a][MOD_b] - PrePSI_GPU[idx][0][0];
                }

                if (a == 0 && b == 0) {
                    LHMAT_GPU[cROW][COL] = PrePSI_GPU[idx][0][0];
                }
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=False)
        SFFTModule_dict['FillLS_PSI'] = _module

        # ************************************ HadProd_PHI.cu & FillLS_PHI.cu ************************************ #

        # ** Hadamard Product [𝚽]
        _refdict = {'N0': N0, 'N1': N1, 'Fpq': Fpq, 'FPHI': FPHI, 'ThreadPerB': MaxThreadPerB}
        _funcstr = r"""
        #include "cuComplex.h"
        extern "C" __global__ void kmain(int SREF_pqp0q0_GPU[%(FPHI)s][2], 
            cuDoubleComplex SPixA_FTpq_GPU[%(Fpq)s][%(N0)s][%(N1)s], 
            cuDoubleComplex SPixA_CFTpq_GPU[%(Fpq)s][%(N0)s][%(N1)s],  
            cuDoubleComplex HpPHI_GPU[%(FPHI)s][%(N0)s][%(N1)s]) 
        {    
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;

            int N0 = %(N0)s;
            int N1 = %(N1)s;
            int Fpq = %(Fpq)s;
            int FPHI = %(FPHI)s;

            __shared__ cuDoubleComplex ShSPixA_FTpq_GPU[%(Fpq)s][%(ThreadPerB)s][%(ThreadPerB)s];
            if (ROW < N0 && COL < N1) {
                for(int pq = 0; pq < Fpq; ++pq){
                    ShSPixA_FTpq_GPU[pq][threadIdx.x][threadIdx.y] = SPixA_FTpq_GPU[pq][ROW][COL];
                }
            }

            __shared__ cuDoubleComplex ShSPixA_CFTpq_GPU[%(Fpq)s][%(ThreadPerB)s][%(ThreadPerB)s];
            if (ROW < N0 && COL < N1) {
                for(int pq = 0; pq < Fpq; ++pq){
                    ShSPixA_CFTpq_GPU[pq][threadIdx.x][threadIdx.y] = SPixA_CFTpq_GPU[pq][ROW][COL];
                }
            }
            __syncthreads();
            
            if (ROW < N0 && COL < N1) {
                for(int p8q8pq = 0; p8q8pq < FPHI; ++p8q8pq){
                    
                    int p8q8 = SREF_pqp0q0_GPU[p8q8pq][0];
                    int pq = SREF_pqp0q0_GPU[p8q8pq][1];

                    HpPHI_GPU[p8q8pq][ROW][COL] = cuCmul(ShSPixA_FTpq_GPU[p8q8][threadIdx.x][threadIdx.y], 
                        ShSPixA_CFTpq_GPU[pq][threadIdx.x][threadIdx.y]);
                }
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=True)
        SFFTModule_dict['HadProd_PHI'] = _module

        # ** Fill Linear-System [𝚽]
        _refdict = {'N0': N0, 'N1': N1, 'NEQ': NEQ, 'Fpq': Fpq, 'Fijab': Fijab, 'FPHI': FPHI}
        _funcstr = r"""
        extern "C" __global__ void kmain(double PrePHI_GPU[%(FPHI)s][%(N0)s][%(N1)s], 
            double LHMAT_GPU[%(NEQ)s][%(NEQ)s]) 
        {
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;
            
            int Fpq = %(Fpq)s;
            int Fijab = %(Fijab)s;

            if (ROW < Fpq && COL < Fpq) {
                
                // INDEX Analysis
                int cROW = Fijab + ROW;              // add offset
                int cCOL = Fijab + COL;              // add offset
                
                int p8q8 = ROW;
                int pq = COL;
                int idx = p8q8 * Fpq + pq;

                // Fill Linear System [C-component]
                LHMAT_GPU[cROW][cCOL] = PrePHI_GPU[idx][0][0];
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=False)
        SFFTModule_dict['FillLS_PHI'] = _module

        # ************************************ HadProd_THE.cu & FillLS_THE.cu ************************************ #

        # ** Hadamard Product [𝚯]
        _refdict = {'N0': N0, 'N1': N1, 'Fij': Fij, 'FTHE': FTHE}
        _funcstr = r"""
        #include "cuComplex.h"
        extern "C" __global__ void kmain(cuDoubleComplex SPixA_FIij_GPU[%(Fij)s][%(N0)s][%(N1)s],
            cuDoubleComplex PixA_CFJ_GPU[%(N0)s][%(N1)s], cuDoubleComplex HpTHE_GPU[%(FTHE)s][%(N0)s][%(N1)s])
        {    
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;

            int N0 = %(N0)s;
            int N1 = %(N1)s;
            int FTHE = %(FTHE)s;

            if (ROW < N0 && COL < N1) {
                for(int i8j8 = 0; i8j8 < FTHE; ++i8j8){
                    HpTHE_GPU[i8j8][ROW][COL] = cuCmul(PixA_CFJ_GPU[ROW][COL], SPixA_FIij_GPU[i8j8][ROW][COL]);
                }
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=True)
        SFFTModule_dict['HadProd_THE'] = _module

        # ** Fill Linear-System [𝚯]
        _refdict = {'N0': N0, 'N1': N1, 'NEQ': NEQ, 'Fijab': Fijab, 'Fab': Fab, 'FTHE': FTHE}
        _funcstr = r"""
        extern "C" __global__ void kmain(int SREF_ijab_GPU[%(Fijab)s][2], int REF_ab_GPU[%(Fab)s][2], 
            double PreTHE_GPU[%(FTHE)s][%(N0)s][%(N1)s], double RHb_GPU[%(NEQ)s]) 
        {
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;
            
            int N0 = %(N0)s;
            int N1 = %(N1)s;
            int Fijab = %(Fijab)s;

            if (ROW < Fijab && COL == 0) {

                // INDEX Analysis
                int i8j8 = SREF_ijab_GPU[ROW][0];
                int a8b8 = SREF_ijab_GPU[ROW][1];
                int a8 = REF_ab_GPU[a8b8][0];
                int b8 = REF_ab_GPU[a8b8][1];
                int idx = i8j8;

                // Define Mod_N0(rho), Mod_N1(eps)
                float tmp = 0.0;
                tmp = fmod(float(a8), float(N0));
                if (tmp < 0.0) {tmp += float(N0);}
                int MODa8 = tmp;

                tmp = fmod(float(b8), float(N1));
                if (tmp < 0.0) {tmp += float(N1);}
                int MODb8 = tmp;
                
                // Fill Linear System [D-component]
                if (a8 != 0 || b8 != 0) {
                    RHb_GPU[ROW] = PreTHE_GPU[idx][MODa8][MODb8] - PreTHE_GPU[idx][0][0];
                }

                if (a8 == 0 && b8 == 0) {
                    RHb_GPU[ROW] = PreTHE_GPU[idx][0][0];
                }
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=False)
        SFFTModule_dict['FillLS_THE'] = _module

        # ************************************ HadProd_DEL.cu & FillLS_DEL.cu ************************************ #

        # ** Hadamard Product [𝚫]
        _refdict = {'N0': N0, 'N1': N1, 'Fpq': Fpq, 'FDEL': FDEL}
        _funcstr = r"""
        #include "cuComplex.h"
        extern "C" __global__ void kmain(cuDoubleComplex SPixA_FTpq_GPU[%(Fpq)s][%(N0)s][%(N1)s], 
            cuDoubleComplex PixA_CFJ_GPU[%(N0)s][%(N1)s], cuDoubleComplex HpDEL_GPU[%(FDEL)s][%(N0)s][%(N1)s])
        {
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;

            int N0 = %(N0)s;
            int N1 = %(N1)s;
            int FDEL = %(FDEL)s;

            if (ROW < N0 && COL < N1) {
                for(int p8q8 = 0; p8q8 < FDEL; ++p8q8){
                    HpDEL_GPU[p8q8][ROW][COL] = 
                        cuCmul(PixA_CFJ_GPU[ROW][COL], SPixA_FTpq_GPU[p8q8][ROW][COL]);
                }
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=True)
        SFFTModule_dict['HadProd_DEL'] = _module

        # ** Fill Linear-System [𝚫]
        _refdict = {'N0': N0, 'N1': N1, 'NEQ': NEQ, 'Fpq': Fpq, 'Fijab': Fijab, 'FDEL': FDEL}
        _funcstr = r"""
        extern "C" __global__ void kmain(double PreDEL_GPU[%(FDEL)s][%(N0)s][%(N1)s], 
            double RHb_GPU[%(NEQ)s]) 
        {
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;
            
            int Fpq = %(Fpq)s;
            int Fijab = %(Fijab)s;

            if (ROW < Fpq && COL == 0) {
                // INDEX Analysis
                int cROW = Fijab + ROW;              // add offset
                int idx = ROW;                       // i.e. p8q8

                // Fill Linear System [E-component]
                RHb_GPU[cROW] = PreDEL_GPU[idx][0][0];
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=False)
        SFFTModule_dict['FillLS_DEL'] = _module

        # ************************************ Remove_LSFStripes.cu ************************************ #

        # ** Remove Forbidden Stripes in Linear-System
        _refdict = {'NEQ': NEQ, 'NEQ_FSfree': NEQ_FSfree}
        _funcstr = r"""
        extern "C" __global__ void kmain(double LHMAT_GPU[%(NEQ)s][%(NEQ)s], int IDX_nFS_GPU[%(NEQ_FSfree)s], 
            double LHMAT_FSfree_GPU[%(NEQ_FSfree)s][%(NEQ_FSfree)s]) 
        {
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;
            int NEQ_FSfree = %(NEQ_FSfree)s;

            if (ROW < NEQ_FSfree && COL < NEQ_FSfree) {
                int ROW_nFS = IDX_nFS_GPU[ROW];
                int COL_nFS = IDX_nFS_GPU[COL];
                LHMAT_FSfree_GPU[ROW][COL] = LHMAT_GPU[ROW_nFS][COL_nFS];
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=False)
        SFFTModule_dict['Remove_LSFStripes'] = _module

        # ************************************ Extend_Solution.cu ************************************ #

        # ** Extend Solution from Forbidden-Stripes-Free Linear-System
        _refdict = {'NEQ_FSfree': NEQ_FSfree, 'NEQ': NEQ}
        _funcstr = r"""
        extern "C" __global__ void kmain(double Solution_FSfree_GPU[%(NEQ_FSfree)s], 
            int IDX_nFS_GPU[%(NEQ_FSfree)s], double Solution_GPU[%(NEQ)s]) 
        {    
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int NEQ_FSfree = %(NEQ_FSfree)s;
            
            if (ROW < NEQ_FSfree) {
                int ROW_nFS = IDX_nFS_GPU[ROW];
                Solution_GPU[ROW_nFS] = Solution_FSfree_GPU[ROW];
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=False)
        SFFTModule_dict['Extend_Solution'] = _module

        # ************************************ Construct_FDIFF.cu ************************************ #

        # ** Construct DIFF Image in Fourier Space 
        _refdict = {'N0': N0, 'N1': N1, 'L0': L0, 'L1': L1, 'w0': w0, 'w1': w1, \
                        'Fijab': Fijab, 'Fab': Fab, 'Fij': Fij, 'Fpq': Fpq, \
                        'SCALE': SCALE, 'ThreadPerB': MaxThreadPerB}
        _funcstr = r"""
        #include "cuComplex.h"
        extern "C" __global__ void kmain(int SREF_ijab_GPU[%(Fijab)s][2], int REF_ab_GPU[%(Fab)s][2],  
            cuDoubleComplex a_ijab_GPU[%(Fijab)s], cuDoubleComplex SPixA_FIij_GPU[%(Fij)s][%(N0)s][%(N1)s], 
            cuDoubleComplex Kab_Wla_GPU[%(L0)s][%(N0)s][%(N1)s], cuDoubleComplex Kab_Wmb_GPU[%(L1)s][%(N0)s][%(N1)s], 
            cuDoubleComplex b_pq_GPU[%(Fpq)s], cuDoubleComplex SPixA_FTpq_GPU[%(Fpq)s][%(N0)s][%(N1)s], 
            cuDoubleComplex PixA_FJ_GPU[%(N0)s][%(N1)s], cuDoubleComplex PixA_FDIFF_GPU[%(N0)s][%(N1)s])
        {       
            int ROW = blockIdx.x*blockDim.x+threadIdx.x;
            int COL = blockIdx.y*blockDim.y+threadIdx.y;
            
            int N0 = %(N0)s;
            int N1 = %(N1)s;
            int w0 = %(w0)s;
            int w1 = %(w1)s;
            int Fij = %(Fij)s;
            int Fab = %(Fab)s;
            int Fpq = %(Fpq)s;

            double One_re = 1.0;
            double Zero_re = 0.0;
            double Zero_im = 0.0;
            double SCALE_re = %(SCALE)s;
            cuDoubleComplex ZERO = make_cuDoubleComplex(Zero_re, Zero_im);
            cuDoubleComplex ONE = make_cuDoubleComplex(One_re, Zero_im);
            cuDoubleComplex SCA = make_cuDoubleComplex(SCALE_re, Zero_im);
            cuDoubleComplex PVAL = ZERO;
            cuDoubleComplex PVAL_FKab = ZERO;

            __shared__ cuDoubleComplex ShSPixA_FIij_GPU[%(Fij)s][%(ThreadPerB)s][%(ThreadPerB)s];
            if (ROW < N0 && COL < N1) {
                for(int ij = 0; ij < Fij; ++ij){
                    ShSPixA_FIij_GPU[ij][threadIdx.x][threadIdx.y] = SPixA_FIij_GPU[ij][ROW][COL];
                }
            }
            __syncthreads();

            if (ROW < N0 && COL < N1) {

                PVAL = ZERO;
                PVAL_FKab = ZERO;
                for(int ab = 0; ab < Fab; ++ab){
                    int a = REF_ab_GPU[ab][0];
                    int b = REF_ab_GPU[ab][1];
                    
                    if (a == 0 && b == 0) {
                        PVAL_FKab = SCA;
                    }
                    
                    if (a != 0 || b != 0) {
                        PVAL_FKab = cuCmul(SCA, cuCsub(cuCmul(Kab_Wla_GPU[w0 + a][ROW][COL], Kab_Wmb_GPU[w1 + b][ROW][COL]), ONE));
                    }

                    for(int ij = 0; ij < Fij; ++ij){
                        int ijab = ij * Fab + ab;
                        PVAL = cuCadd(PVAL, cuCmul(cuCmul(a_ijab_GPU[ijab], ShSPixA_FIij_GPU[ij][threadIdx.x][threadIdx.y]), PVAL_FKab));
                    }
                }

                for(int pq = 0; pq < Fpq; ++pq){
                    PVAL = cuCadd(PVAL, cuCmul(b_pq_GPU[pq], SPixA_FTpq_GPU[pq][ROW][COL]));
                }
                
                PixA_FDIFF_GPU[ROW][COL] = cuCsub(PixA_FJ_GPU[ROW][COL], PVAL);
            }
        }
        """
        _code = _funcstr % _refdict
        _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=True)
        SFFTModule_dict['Construct_FDIFF'] = _module
        
        SFFTConfig = (SFFTParam_dict, SFFTModule_dict)
        if VERBOSE_LEVEL in [1, 2]:
            print('\n --//--//--//--//-- EXIT SFFT COMPILATION --//--//--//--//-- ')

        return SFFTConfig

class SingleSFFTConfigure_Numpy:
    @staticmethod
    def SSCN(NX, NY, KerHW, KerPolyOrder=2, BGPolyOrder=2, ConstPhotRatio=True, 
             NUM_CPU_THREADS_4SUBTRACT=8, NUMBA_CACHE=True, VERBOSE_LEVEL=2):

        import numba as nb
        nb.set_num_threads = NUM_CPU_THREADS_4SUBTRACT

        N0, N1 = int(NX), int(NY)
        w0, w1 = int(KerHW), int(KerHW)
        DK, DB = int(KerPolyOrder), int(BGPolyOrder)
        SFFTParam_dict = {}
        SFFTModule_dict = {}
        
        if DK not in [0, 1, 2, 3]:
            # NOTE DK = 0 means a spatial invariant matching-kernel.
            raise Exception('MeLOn ERROR: Input KerPolyOrder should be 0/1/2/3!')
        
        if DB not in [0, 1, 2, 3]:
            # NOTE DB = 0 means a flat differential-background, however, we cannot kill the sky term in our script.
            raise Exception('MeLOn ERROR: Input BGPolyOrder should be 0/1/2/3!')

        if VERBOSE_LEVEL in [1, 2]:
            print('\n --//--//--//--//-- TRIGGER SFFT COMPILATION --//--//--//--//-- ')
            print('\n ---//--- KerPolyOrder %d | BGPolyOrder %d | KerHW [%d] ---//--- ' %(DK, DB, w0))

        # * Make a dictionary for SFFT parameters
        L0 = 2*w0+1                       # matching-kernel XSize
        L1 = 2*w1+1                       # matching-kernel YSize
        Fab = L0 * L1                     # dof for index 𝛼𝛽
        Fij = int((DK+1)*(DK+2)/2)        # dof for matching-kernel polynomial index ij 
        Fpq = int((DB+1)*(DB+2)/2)        # dof for differential-background polynomial index pq 
        SCALE = np.float64(1/(N0*N1))     # Scale of Image-Size
        SCALE_L = np.float64(1/SCALE)     # Reciprocal Scale of Image Size

        NEQ = Fij*Fab+Fpq                         # Linear-System side-length
        Fijab = Fij * Fab                         # Linear-System Major side-length
        NEQ_FSfree = NEQ - (Fij - 1)              # Forbidden-Stripes-Free Linear-System side-length

        FOMG, FGAM, FTHE = Fij**2, Fij*Fpq, Fij   # OMG / GAM / THE has shape (dof, N0, N1)
        FPSI, FPHI, FDEL = Fpq*Fij, Fpq**2, Fpq   # PSI / PHI / DEL has shape (dof, N0, N1)

        SFFTParam_dict['N0'] = N0
        SFFTParam_dict['N1'] = N1
        SFFTParam_dict['w0'] = w0
        SFFTParam_dict['w1'] = w1
        SFFTParam_dict['DK'] = DK
        SFFTParam_dict['DB'] = DB
        SFFTParam_dict['ConstPhotRatio'] = ConstPhotRatio

        SFFTParam_dict['L0'] = L0
        SFFTParam_dict['L1'] = L1
        SFFTParam_dict['Fab'] = Fab
        SFFTParam_dict['Fij'] = Fij
        SFFTParam_dict['Fpq'] = Fpq
        SFFTParam_dict['SCALE'] = SCALE
        SFFTParam_dict['SCALE_L'] = SCALE_L
        SFFTParam_dict['NEQ'] = NEQ
        SFFTParam_dict['Fijab'] = Fijab
        SFFTParam_dict['NEQ_FSfree'] = NEQ_FSfree

        SFFTParam_dict['FOMG'] = FOMG
        SFFTParam_dict['FGAM'] = FGAM
        SFFTParam_dict['FTHE'] = FTHE
        SFFTParam_dict['FPSI'] = FPSI
        SFFTParam_dict['FPHI'] = FPHI
        SFFTParam_dict['FDEL'] = FDEL

        # * Load SFFT Numpy-based modules
        # ************************************ SpatialCoor.py ************************************ #

        # ** produce spatial coordinate X/Y/oX/oY-map
        _strdec = 'Tuple((i4[:,:], i4[:,:], f8[:,:], f8[:,:]))(i4[:,:], i4[:,:], f8[:,:], f8[:,:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def SpatialCoor(PixA_X, PixA_Y, PixA_CX, PixA_CY):

            #assert PixA_X.dtype == np.int32 and PixA_X.shape == (N0, N1)
            #assert PixA_Y.dtype == np.int32 and PixA_Y.shape == (N0, N1)
            #assert PixA_CX.dtype == np.float64 and PixA_CX.shape == (N0, N1)
            #assert PixA_CY.dtype == np.float64 and PixA_CY.shape == (N0, N1)

            for ROW in nb.prange(N0):
                for COL in nb.prange(N1):
                    PixA_X[ROW, COL] = ROW
                    PixA_Y[ROW, COL] = COL
                    PixA_CX[ROW, COL] = (float(ROW) + 1.0) / N0   # NOTE UPDATE
                    PixA_CY[ROW, COL] = (float(COL) + 1.0) / N1   # NOTE UPDATE

            return PixA_X, PixA_Y, PixA_CX, PixA_CY

        SFFTModule_dict['SpatialCoor'] = SpatialCoor
        # ************************************ SpatialPoly.py ************************************ #

        # ** produce Iij, Tpq
        _strdec = 'Tuple((f8[:,:,:], f8[:,:,:]))' + '(i4[:,:], i4[:,:], f8[:,:], f8[:,:], f8[:,:], f8[:,:,:], f8[:,:,:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def SpatialPoly(REF_ij, REF_pq, PixA_CX, PixA_CY, PixA_I, SPixA_Iij, SPixA_Tpq):
            
            #assert REF_ij.dtype == np.int32 and REF_ij.shape == (Fij, 2)
            #assert REF_pq.dtype == np.int32 and REF_pq.shape == (Fpq, 2)
            
            #assert PixA_CX.dtype == np.float64 and PixA_CX.shape == (N0, N1)
            #assert PixA_CY.dtype == np.float64 and PixA_CY.shape == (N0, N1)
            #assert PixA_I.dtype == np.float64 and PixA_I.shape == (N0, N1)
            
            #assert SPixA_Iij.dtype == np.float64 and SPixA_Iij.shape == (Fij, N0, N1)
            #assert SPixA_Tpq.dtype == np.float64 and SPixA_Tpq.shape == (Fpq, N0, N1)

            for ij in nb.prange(Fij):
                i, j = REF_ij[ij]
                PixA_kpoly = np.power(PixA_CX, i) * np.power(PixA_CY, j)
                SPixA_Iij[ij, :, :] = PixA_I * PixA_kpoly   # NOTE UPDATE
            
            for pq in nb.prange(Fpq):
                p, q = REF_pq[pq]
                PixA_bpoly = np.power(PixA_CX, p) * np.power(PixA_CY, q)
                SPixA_Tpq[pq, :, :] = PixA_bpoly   # NOTE UPDATE
            
            return SPixA_Iij, SPixA_Tpq

        SFFTModule_dict['SpatialPoly'] = SpatialPoly
        # ************************************ HadProd_OMG.py & FillLS_OMG.py ************************************ #

        # ** Hadamard Product [𝛀]
        _strdec = 'c16[:,:,:](i4[:,:], c16[:,:,:], c16[:,:,:], c16[:,:,:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def HadProd_OMG(SREF_iji0j0, SPixA_FIij, SPixA_CFIij, HpOMG):
            
            #assert SREF_iji0j0.dtype == np.int32 and SREF_iji0j0.shape == (FOMG, 2)
            #assert SPixA_FIij.dtype == np.complex128 and SPixA_FIij.shape == (Fij, N0, N1)
            #assert SPixA_CFIij.dtype == np.complex128 and SPixA_CFIij.shape == (Fij, N0, N1)
            #assert HpOMG.dtype == np.complex128 and HpOMG.shape == (FOMG, N0, N1)

            for i8j8ij in nb.prange(FOMG):
                i8j8, ij = SREF_iji0j0[i8j8ij]
                HpOMG[i8j8ij, :, :] = SPixA_FIij[i8j8] * SPixA_CFIij[ij]   # NOTE UPDATE

            return HpOMG

        # ** Fill Linear-System [𝛀]
        _strdec = 'f8[:,:](i4[:,:], i4[:,:], f8[:,:,:], f8[:,:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def FillLS_OMG(SREF_ijab, REF_ab, PreOMG, LHMAT):
            
            #assert SREF_ijab.dtype == np.int32 and SREF_ijab.shape == (Fijab, 2)
            #assert REF_ab.dtype == np.int32 and REF_ab.shape == (Fab, 2)
            #assert PreOMG.dtype == np.float64 and PreOMG.shape == (FOMG, N0, N1)
            #assert LHMAT.dtype == np.float64 and LHMAT.shape == (NEQ, NEQ)
            
            for ROW in nb.prange(Fijab):
                for COL in nb.prange(Fijab):
                    
                    # *** analysze index
                    i8j8, a8b8 = SREF_ijab[ROW]
                    ij, ab = SREF_ijab[COL]

                    a8, b8 = REF_ab[a8b8]
                    a, b = REF_ab[ab]
                    idx = i8j8 * Fij + ij
                    
                    # *** define Mod_N0(rho), Mod_N1(eps)
                    tmp = int(np.fmod(float(a8), float(N0)))
                    if tmp < 0.0: tmp += N0
                    MODa8 = tmp
                    
                    tmp = int(np.fmod(float(b8), float(N1)))
                    if tmp < 0.0: tmp += N1
                    MODb8 = tmp

                    tmp = int(np.fmod(float(-a), float(N0)))
                    if tmp < 0.0: tmp += N0
                    MOD_a = tmp

                    tmp = int(np.fmod(float(-b), float(N1)))
                    if tmp < 0.0: tmp += N1
                    MOD_b = tmp

                    tmp = int(np.fmod(float(a8-a), float(N0)))
                    if tmp < 0.0: tmp += N0
                    MODa8_a = tmp

                    tmp = int(np.fmod(float(b8-b), float(N1)))
                    if tmp < 0.0: tmp += N1
                    MODb8_b = tmp
                
                    # *** fill linear system [A-component]
                    if ((a8 != 0 or b8 != 0) and (a != 0 or b != 0)):
                        LHMAT[ROW, COL] = - PreOMG[idx, MODa8, MODb8] \
                                          - PreOMG[idx, MOD_a, MOD_b] \
                                          + PreOMG[idx, MODa8_a, MODb8_b] \
                                          + PreOMG[idx, 0, 0]   # NOTE UPDATE

                    if ((a8 == 0 and b8 == 0) and (a != 0 or b != 0)):
                        LHMAT[ROW, COL] = PreOMG[idx, MOD_a, MOD_b] - PreOMG[idx, 0, 0]   # NOTE UPDATE

                    if ((a8 != 0 or b8 != 0) and (a == 0 and b == 0)):
                        LHMAT[ROW, COL] = PreOMG[idx, MODa8, MODb8] - PreOMG[idx, 0, 0]   # NOTE UPDATE

                    if ((a8 == 0 and b8 == 0) and (a == 0 and b == 0)):
                        LHMAT[ROW, COL] = PreOMG[idx, 0, 0]   # NOTE UPDATE
                    
            return LHMAT

        SFFTModule_dict['HadProd_OMG'] = HadProd_OMG
        SFFTModule_dict['FillLS_OMG'] = FillLS_OMG
        # ************************************ HadProd_GAM.py & FillLS_GAM.py ************************************ #

        # ** Hadamard Product [𝜦]
        _strdec = 'c16[:,:,:](i4[:,:], c16[:,:,:], c16[:,:,:], c16[:,:,:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def HadProd_GAM(SREF_ijpq, SPixA_FIij, SPixA_CFTpq, HpGAM):
            
            #assert SREF_ijpq.dtype == np.int32 and SREF_ijpq.shape == (FGAM, 2)
            #assert SPixA_FIij.dtype == np.complex128 and SPixA_FIij.shape == (Fij, N0, N1)
            #assert SPixA_CFTpq.dtype == np.complex128 and SPixA_CFTpq.shape == (Fpq, N0, N1)
            #assert HpGAM.dtype == np.complex128 and HpGAM.shape == (FGAM, N0, N1)
            
            for i8j8pq in nb.prange(FGAM):
                i8j8, pq = SREF_ijpq[i8j8pq]
                HpGAM[i8j8pq, :, :] = SPixA_FIij[i8j8] * SPixA_CFTpq[pq]   # NOTE UPDATE

            return HpGAM

        # ** Fill Linear-System [𝜦]
        _strdec = 'f8[:,:](i4[:,:], i4[:,:], f8[:,:,:], f8[:,:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def FillLS_GAM(SREF_ijab, REF_ab, PreGAM, LHMAT):
            
            #assert SREF_ijab.dtype == np.int32 and SREF_ijab.shape == (Fijab, 2)
            #assert REF_ab.dtype == np.int32 and REF_ab.shape == (Fab, 2)
            #assert PreGAM.dtype == np.float64 and PreGAM.shape == (FGAM, N0, N1)
            #assert LHMAT.dtype == np.float64 and LHMAT.shape == (NEQ, NEQ)
            
            for ROW in nb.prange(Fijab):
                for COL in nb.prange(Fpq):
                    
                    # *** analysze index
                    i8j8, a8b8 = SREF_ijab[ROW]
                    pq = COL
                    
                    a8, b8 = REF_ab[a8b8]
                    idx = i8j8 * Fpq + pq
                    cCOL = Fijab + COL       # add offset
                    
                    # *** define Mod_N0(rho), Mod_N1(eps)
                    tmp = int(np.fmod(float(a8), float(N0)))
                    if tmp < 0.0: tmp += N0
                    MODa8 = tmp

                    tmp = int(np.fmod(float(b8), float(N1)))
                    if tmp < 0.0: tmp += N1
                    MODb8 = tmp
                    
                    # *** fill linear system [B-component]
                    if (a8 != 0 or b8 != 0):
                        LHMAT[ROW, cCOL] = PreGAM[idx, MODa8, MODb8] - PreGAM[idx, 0, 0]   # NOTE UPDATE
                    
                    if (a8 == 0 and b8 == 0):
                        LHMAT[ROW, cCOL] = PreGAM[idx, 0, 0]   # NOTE UPDATE
            
            return LHMAT

        SFFTModule_dict['HadProd_GAM'] = HadProd_GAM
        SFFTModule_dict['FillLS_GAM'] = FillLS_GAM
        # ************************************ HadProd_PSI.py & FillLS_PSI.py ************************************ #

        # ** Hadamard Product [𝜳]
        _strdec = 'c16[:,:,:](i4[:,:], c16[:,:,:], c16[:,:,:], c16[:,:,:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def HadProd_PSI(SREF_pqij, SPixA_CFIij, SPixA_FTpq, HpPSI):
            
            #assert SREF_pqij.dtype == np.int32 and SREF_pqij.shape == (FPSI, 2)
            #assert SPixA_CFIij.dtype == np.complex128 and SPixA_CFIij.shape == (Fij, N0, N1)
            #assert SPixA_FTpq.dtype == np.complex128 and SPixA_FTpq.shape == (Fpq, N0, N1)
            #assert HpPSI.dtype == np.complex128 and HpPSI.shape == (FPSI, N0, N1)

            for p8q8ij in nb.prange(FPSI):
                p8q8, ij = SREF_pqij[p8q8ij]
                HpPSI[p8q8ij, :, :] = SPixA_FTpq[p8q8] * SPixA_CFIij[ij]   # NOTE UPDATE

            return HpPSI

        # ** Fill Linear-System [𝜳]
        _strdec = 'f8[:,:](i4[:,:], i4[:,:], f8[:,:,:], f8[:,:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def FillLS_PSI(SREF_ijab, REF_ab, PrePSI, LHMAT):
            
            #assert SREF_ijab.dtype == np.int32 and SREF_ijab.shape == (Fijab, 2)
            #assert REF_ab.dtype == np.int32 and REF_ab.shape == (Fab, 2)
            #assert PrePSI.dtype == np.float64 and PrePSI.shape == (FPSI, N0, N1)
            #assert LHMAT.dtype == np.float64 and LHMAT.shape == (NEQ, NEQ)

            for ROW in nb.prange(Fpq):
                for COL in nb.prange(Fijab):
                    
                    # *** analysze index
                    cROW = Fijab + ROW       # add offset
                    p8q8 = ROW
                    
                    ij, ab = SREF_ijab[COL]
                    a, b = REF_ab[ab]
                    idx = p8q8 * Fij + ij
                    
                    # *** define Mod_N0(rho), Mod_N1(eps)
                    tmp = int(np.fmod(float(-a), float(N0)))
                    if tmp < 0.0: tmp += N0
                    MOD_a = tmp

                    tmp = int(np.fmod(float(-b), float(N1)))
                    if tmp < 0.0: tmp += N1
                    MOD_b = tmp
                    
                    # *** fill linear system [B#-component]
                    if (a != 0 or b != 0):
                        LHMAT[cROW, COL] = PrePSI[idx, MOD_a, MOD_b] - PrePSI[idx, 0, 0]   # NOTE UPDATE

                    if (a == 0 and b == 0):
                        LHMAT[cROW, COL] = PrePSI[idx, 0, 0]   # NOTE UPDATE
                    
            return LHMAT

        SFFTModule_dict['HadProd_PSI'] = HadProd_PSI
        SFFTModule_dict['FillLS_PSI'] = FillLS_PSI
        # ************************************ HadProd_PHI.py & FillLS_PHI.py ************************************ #

        # ** Hadamard Product [𝚽]
        _strdec = 'c16[:,:,:](i4[:,:], c16[:,:,:], c16[:,:,:], c16[:,:,:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def HadProd_PHI(SREF_pqp0q0, SPixA_FTpq, SPixA_CFTpq, HpPHI):
            
            #assert SREF_pqp0q0.dtype == np.int32 and SREF_pqp0q0.shape == (FPHI, 2)
            #assert SPixA_FTpq.dtype == np.complex128 and SPixA_FTpq.shape == (Fpq, N0, N1)
            #assert SPixA_CFTpq.dtype == np.complex128 and SPixA_CFTpq.shape == (Fpq, N0, N1)
            #assert HpPHI.dtype == np.complex128 and HpPHI.shape == (FPHI, N0, N1)
            
            for p8q8pq in nb.prange(FPHI):
                p8q8, pq = SREF_pqp0q0[p8q8pq]
                HpPHI[p8q8pq, :, :] = SPixA_FTpq[p8q8] * SPixA_CFTpq[pq]   # NOTE UPDATE

            return HpPHI

        # ** Fill Linear-System [𝚽]
        _strdec = 'f8[:,:](f8[:,:,:], f8[:,:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def FillLS_PHI(PrePHI, LHMAT):
            
            #assert PrePHI.dtype == np.float64 and PrePHI.shape == (FPHI, N0, N1)
            #assert LHMAT.dtype == np.float64 and LHMAT.shape == (NEQ, NEQ)
            
            for ROW in nb.prange(Fpq):
                for COL in nb.prange(Fpq):
                    
                    # *** analysze index
                    cROW = Fijab + ROW       # add offset
                    cCOL = Fijab + COL       # add offset
                    
                    p8q8 = ROW
                    pq = COL
                    idx = p8q8 * Fpq + pq
                    
                    # *** fill linear system [C-component]
                    LHMAT[cROW, cCOL] = PrePHI[idx, 0, 0]   # NOTE UPDATE

            return LHMAT

        SFFTModule_dict['HadProd_PHI'] = HadProd_PHI
        SFFTModule_dict['FillLS_PHI'] = FillLS_PHI
        # ************************************ HadProd_THE.py & FillLS_THE.py ************************************ #

        # ** Hadamard Product [𝚯]
        _strdec = 'c16[:,:,:](c16[:,:,:], c16[:,:], c16[:,:,:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def HadProd_THE(SPixA_FIij, PixA_CFJ, HpTHE):
            
            #assert SPixA_FIij.dtype == np.complex128 and SPixA_FIij.shape == (Fij, N0, N1)
            #assert PixA_CFJ.dtype == np.complex128 and PixA_CFJ.shape == (N0, N1)
            #assert HpTHE.dtype == np.complex128 and HpTHE.shape == (FTHE, N0, N1)
            
            for i8j8 in nb.prange(FTHE):
                HpTHE[i8j8, :, :] = PixA_CFJ * SPixA_FIij[i8j8]   # NOTE UPDATE

            return HpTHE

        # ** Fill Linear-System [𝚯]
        _strdec = 'f8[:](i4[:,:], i4[:,:], f8[:,:,:], f8[:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def FillLS_THE(SREF_ijab, REF_ab, PreTHE, RHb):
            
            #assert SREF_ijab.dtype == np.int32 and SREF_ijab.shape == (Fijab, 2)
            #assert REF_ab.dtype == np.int32 and REF_ab.shape == (Fab, 2)
            #assert PreTHE.dtype == np.float64 and PreTHE.shape == (FTHE, N0, N1)
            #assert RHb.dtype == np.float64 and RHb.shape == (NEQ, )
            
            COL = 0   # trivial
            for ROW in nb.prange(Fijab):
                
                # *** analysze index
                i8j8, a8b8 = SREF_ijab[ROW]
                a8, b8 = REF_ab[a8b8]
                idx = i8j8
                
                # *** define Mod_N0(rho), Mod_N1(eps)
                tmp = int(np.fmod(float(a8), float(N0)))
                if tmp < 0.0: tmp += N0
                MODa8 = tmp

                tmp = int(np.fmod(float(b8), float(N1)))
                if tmp < 0.0: tmp += N1
                MODb8 = tmp
                
                # *** fill linear system [D-component]
                if (a8 != 0 or b8 != 0):
                    RHb[ROW] = PreTHE[idx, MODa8, MODb8] - PreTHE[idx, 0, 0]   # NOTE UPDATE

                if (a8 == 0 and b8 == 0):
                    RHb[ROW] = PreTHE[idx, 0, 0]   # NOTE UPDATE
            
            return RHb

        SFFTModule_dict['HadProd_THE'] = HadProd_THE
        SFFTModule_dict['FillLS_THE'] = FillLS_THE
        # ************************************ HadProd_DEL.py & FillLS_DEL.py ************************************ #

        # ** Hadamard Product [𝚫]
        _strdec = 'c16[:,:,:](c16[:,:,:], c16[:,:], c16[:,:,:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def HadProd_DEL(SPixA_FTpq, PixA_CFJ, HpDEL):
            
            #assert SPixA_FTpq.dtype == np.complex128 and SPixA_FTpq.shape == (Fpq, N0, N1)
            #assert PixA_CFJ.dtype == np.complex128 and PixA_CFJ.shape == (N0, N1)
            #assert HpDEL.dtype == np.complex128 and HpDEL.shape == (FDEL, N0, N1)
            
            for p8q8 in nb.prange(FDEL):
                HpDEL[p8q8, :, :] = PixA_CFJ * SPixA_FTpq[p8q8]   # NOTE UPDATE

            return HpDEL

        # ** Fill Linear-System [𝚫]
        _strdec = 'f8[:](f8[:,:,:], f8[:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def FillLS_DEL(PreDEL, RHb):
            
            #assert PreDEL.dtype == np.float64 and PreDEL.shape == (FDEL, N0, N1)
            #assert RHb.dtype == np.float64 and RHb.shape == (NEQ, )
            
            COL = 0   # trivial
            for ROW in nb.prange(Fpq):
                
                # *** analysze index
                cROW = Fijab + ROW       # add offset
                idx = ROW                # i.e. p8q8
                
                # *** fill linear system [E-component]
                RHb[cROW] = PreDEL[idx, 0, 0]   # NOTE UPDATE
            
            return RHb

        SFFTModule_dict['HadProd_DEL'] = HadProd_DEL
        SFFTModule_dict['FillLS_DEL'] = FillLS_DEL
        # ************************************ Remove_LSFStripes.py ************************************ #

        # ** Remove Forbidden Stripes in Linear-System
        _strdec = 'f8[:,:](f8[:,:], i4[:], f8[:,:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def Remove_LSFStripes(LHMAT, IDX_nFS, LHMAT_FSfree):
            
            #assert LHMAT.dtype == np.float64 and LHMAT.shape == (NEQ, NEQ)
            #assert IDX_nFS.dtype == np.int32 and IDX_nFS.shape == (NEQ_FSfree, )
            #assert LHMAT_FSfree.dtype == np.float64 and LHMAT_FSfree.shape == (NEQ_FSfree, NEQ_FSfree)
            
            for ROW in nb.prange(NEQ_FSfree):
                for COL in nb.prange(NEQ_FSfree):
                    ROW_nFS = IDX_nFS[ROW]
                    COL_nFS = IDX_nFS[COL]
                    LHMAT_FSfree[ROW, COL] = LHMAT[ROW_nFS, COL_nFS]   # NOTE UPDATE
            
            return LHMAT_FSfree

        SFFTModule_dict['Remove_LSFStripes'] = Remove_LSFStripes
        # ************************************ Extend_Solution.py ************************************ #

        # ** Extend Solution from Forbidden-Stripes-Free Linear-System
        _strdec = 'f8[:](f8[:], i4[:], f8[:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def Extend_Solution(Solution_FSfree, IDX_nFS, Solution):
            
            #assert Solution_FSfree.dtype == np.float64 and Solution_FSfree.shape == (NEQ_FSfree, )
            #assert IDX_nFS.dtype == np.int32 and IDX_nFS.shape == (NEQ_FSfree, )
            #assert Solution.dtype == np.float64 and Solution.shape == (NEQ, )
            
            for ROW in nb.prange(NEQ_FSfree):
                ROW_nFS = IDX_nFS[ROW]
                Solution[ROW_nFS] = Solution_FSfree[ROW]   # NOTE UPDATE
            
            return Solution

        SFFTModule_dict['Extend_Solution'] = Extend_Solution
        # ************************************ Construct_FDIFF.py ************************************ #

        _strdec = 'c16[:,:](i4[:,:], i4[:,:], c16[:], c16[:,:,:], c16[:,:,:], c16[:,:,:], c16[:], c16[:,:,:], c16[:,:], c16[:,:])'
        @nb.njit(_strdec, parallel=True, cache=NUMBA_CACHE)
        def Construct_FDIFF(SREF_ijab, REF_ab, a_ijab, SPixA_FIij, Kab_Wla, Kab_Wmb, b_pq, SPixA_FTpq, PixA_FJ, PixA_FDIFF):
            
            #assert SREF_ijab.dtype == np.int32 and SREF_ijab.shape == (Fijab, 2)
            #assert REF_ab.dtype == np.int32 and REF_ab.shape == (Fab, 2)
            
            #assert a_ijab.dtype == np.complex128 and a_ijab.shape == (Fijab, )
            #assert SPixA_FIij.dtype == np.complex128 and SPixA_FIij.shape == (Fij, N0, N1)
            #assert Kab_Wla.dtype == np.complex128 and Kab_Wla.shape == (L0, N0, N1)
            #assert Kab_Wmb.dtype == np.complex128 and Kab_Wmb.shape == (L1, N0, N1)
            #assert b_pq.dtype == np.complex128 and b_pq.shape == (Fpq, )
            #assert SPixA_FTpq.dtype == np.complex128 and SPixA_FTpq.shape == (Fpq, N0, N1)
            
            #assert PixA_FJ.dtype == np.complex128 and PixA_FJ.shape == (N0, N1)
            #assert PixA_FDIFF.dtype == np.complex128 and PixA_FDIFF.shape == (N0, N1)

            ZERO_C = 0.0 + 0.0j
            ONE_C = 1.0 + 0.0j
            SCALE_C = SCALE + 0.0j

            for ROW in nb.prange(N0):
                for COL in nb.prange(N1):
                    
                    PVAL = ZERO_C
                    for ab in nb.prange(Fab):
                        a, b = REF_ab[ab]

                        if (a == 0 and b == 0):
                            PVAL_FKab = SCALE_C

                        if (a != 0 or b != 0):
                            PVAL_FKab = SCALE_C * ((Kab_Wla[w0 + a, ROW, COL] * Kab_Wmb[w1 + b, ROW, COL]) - ONE_C)

                        for ij in nb.prange(Fij):
                            ijab = ij * Fab + ab
                            PVAL += (a_ijab[ijab] * SPixA_FIij[ij, ROW, COL]) * PVAL_FKab
                        
                    for pq in nb.prange(Fpq):
                        PVAL += b_pq[pq] * SPixA_FTpq[pq, ROW, COL]

                    PixA_FDIFF[ROW, COL] = PixA_FJ[ROW, COL] - PVAL
            
            return PixA_FDIFF

        SFFTModule_dict['Construct_FDIFF'] = Construct_FDIFF

        SFFTConfig = (SFFTParam_dict, SFFTModule_dict)
        if VERBOSE_LEVEL in [1, 2]: 
            print('\n --//--//--//--//-- EXIT SFFT COMPILATION --//--//--//--//-- ')

        return SFFTConfig

class SingleSFFTConfigure:
    @staticmethod
    def SSC(NX, NY, KerHW, KerPolyOrder=2, BGPolyOrder=2, ConstPhotRatio=True, \
        BACKEND_4SUBTRACT='Cupy', NUM_CPU_THREADS_4SUBTRACT=8, NUMBA_CACHE=True, VERBOSE_LEVEL=2):

        """
        # Arguments:
        # a) NX, NY: Image Size (pixels) along X / Y axis.
        # b) KerHW: Kernel half width for compilation.
        # c) KerPolyOrder: The order of Polynomial Variation for PSF-Matching Kernel.
        # d) BGPolyOrder: The order of Polynomial Variation for Differential Background.
        # e) ConstPhotRatio: Use a constant photometric ratio in image subtraction?
        # f) BACKEND_4SUBTRACT : Which backend would you like to perform SFFT subtraction on? (Cupy/Numpy)
        # h) NUM_CPU_THREADS_4SUBTRACT: How many CPU threads would you want to perform SFFT subtraction on? (Numpy)
        # i) NUMBA_CACHE: Whether to cache the compiled Numba functions into NUMBA_CACHE_DIR or not.
        #
        """
        
        if BACKEND_4SUBTRACT == 'Cupy':
            SFFTConfig = SingleSFFTConfigure_Cupy.SSCC(NX=NX, NY=NY, KerHW=KerHW, \
                KerPolyOrder=KerPolyOrder, BGPolyOrder=BGPolyOrder, ConstPhotRatio=ConstPhotRatio, \
                VERBOSE_LEVEL=VERBOSE_LEVEL)
        
        if BACKEND_4SUBTRACT == 'Numpy':
            SFFTConfig = SingleSFFTConfigure_Numpy.SSCN(NX=NX, NY=NY, KerHW=KerHW, \
                KerPolyOrder=KerPolyOrder, BGPolyOrder=BGPolyOrder, ConstPhotRatio=ConstPhotRatio, \
                NUM_CPU_THREADS_4SUBTRACT=NUM_CPU_THREADS_4SUBTRACT, NUMBA_CACHE=NUMBA_CACHE, VERBOSE_LEVEL=VERBOSE_LEVEL)
        
        return SFFTConfig



import importlib.util as iu
if iu.find_spec('cupy') is not None:
    import cupy as cp

__last_update__ = "2024-09-22"
__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.6.1"

#########################################################################
#||                                                                   ||#
#||                           Acknowledgment                          ||#
#||                                                                   ||#
#########################################################################
#||                                                                   ||#
#||   This code was developed during the NASA GPU Hackathon 2024      ||#
#||   (https://www.nas.nasa.gov/hackathon/#home) as part of the       ||#
#||   pipeline optimization efforts for the Roman Supernova PIT       ||#
#||   team (https://github.com/Roman-Supernova-PIT). I extend my      ||#
#||   sincere appreciation to my team members and our Hackathon       ||#
#||   mentors for their valuable support and insightful               ||#
#||   suggestions.                                                    ||#
#||                                                                   ||#
#||   I would like to specifically acknowledge the contributions      ||#
#||   of the following team members: Lauren Aldoroty (Duke),          ||#
#||   Robert Knop (LBNL), Shu Liu (Pitt), and Michael Wood-Vasey      ||#
#||   (Pitt). Additionally, I am grateful for the guidance provided   ||#
#||   by our mentors, Marcus Manos and Lucas Erlandson from NVIDIA.   ||#
#||                                                                   ||#
#||   The collaborative environment fostered by the NASA GPU          ||#
#||   Hackathon and the Roman Supernova PIT team was essential for    ||#
#||   the development and optimization of this code.                  ||#
#||                                                                   ||#
#########################################################################

class PureCupy_FFTKits:
    @staticmethod
    def KERNEL_CSZ(KERNEL_GPU, NX_IMG, NY_IMG, NORMALIZE_KERNEL=False):
        """ Circular Shift the kernel and extend to the target size """
        N0, N1 = NX_IMG, NY_IMG
        L0, L1 = KERNEL_GPU.shape
        W0, W1 = (L0 - 1)//2, (L1 - 1)//2
        # Note: currently only support odd-sized kernel
        assert L0 % 2 == 1 and L1 % 2 == 1   
        
        if NORMALIZE_KERNEL:
            KERNEL_TZP_GPU = cp.pad(KERNEL_GPU / cp.sum(KERNEL_GPU), \
                pad_width=((0, N0 - L0), (0, N1 - L1)), mode='constant', constant_values=0.)
        else:   
            KERNEL_TZP_GPU = cp.pad(KERNEL_GPU, pad_width=((0, N0 - L0), (0, N1 - L1)), \
                mode='constant', constant_values=0.)
        KIMG_CSZ_GPU = cp.roll(cp.roll(KERNEL_TZP_GPU, -W0, axis=0), -W1, axis=1)
        return KIMG_CSZ_GPU

    @staticmethod
    def KERNEL_CSZ_INV(KIMG_GPU, NX_KERN, NY_KERN, VERBOSE_LEVEL=2):
        """ Inverse Circular Shift the kernel and truncate to the target size """
        L0, L1 = NX_KERN, NY_KERN
        W0, W1 = (L0 - 1)//2, (L1 - 1)//2
        # Note: currently only support odd-sized kernel
        assert L0 % 2 == 1 and L1 % 2 == 1
        
        KIMG_iCSZ_GPU = cp.roll(cp.roll(KIMG_GPU, W1, axis=1), W0, axis=0)
        KERNEL_GPU = KIMG_iCSZ_GPU[:L0, :L1]
        if VERBOSE_LEVEL in [1, 2]:
            LOSE_RATIO = 1. - cp.sum(cp.abs(KERNEL_GPU)) / cp.sum(cp.abs(KIMG_iCSZ_GPU))
            _report_message = "Kernel Truncation Loses APE = [%.4f %s]" %(LOSE_RATIO*100, '%')
            print("MeLOn CheckPoint: %s " % _report_message)
        return KERNEL_GPU

    @staticmethod
    def FFT_CONVOLVE(PixA_Inp_GPU, KERNEL_GPU, PAD_FILL_VALUE=0., NAN_FILL_VALUE=0., 
        NORMALIZE_KERNEL=False, FORCE_OUTPUT_C_CONTIGUOUS=False, FFT_BACKEND="Cupy"):
        """ FFT Convolition """
        N0, N1 = PixA_Inp_GPU.shape
        L0, L1 = KERNEL_GPU.shape
        assert L0 % 2 == 1 and L1 % 2 == 1

        W0, W1 = (L0 - 1)//2, (L1 - 1)//2 
        NX_IMG, NY_IMG = N0 + 2*W0, N1 + 2*W1

        # zero padding on input image
        PixA_EInp_GPU = cp.pad(
            array=PixA_Inp_GPU, 
            pad_width=((W0, W0), (W1, W1)), 
            mode='constant', 
            constant_values=PAD_FILL_VALUE
        )

        if NAN_FILL_VALUE is not None:
            PixA_EInp_GPU[cp.isnan(PixA_EInp_GPU)] = NAN_FILL_VALUE
        
        # circular shift the kernel to the center and extend to the target size
        KIMG_CSZ_GPU = PureCupy_FFTKits.KERNEL_CSZ(KERNEL_GPU=KERNEL_GPU, 
            NX_IMG=NX_IMG, NY_IMG=NY_IMG, NORMALIZE_KERNEL=NORMALIZE_KERNEL)

        # perform convolution in Fourier domain
        if FFT_BACKEND == "Cupy":
            PixA_Out_GPU = cp.fft.ifft2(
                cp.fft.fft2(PixA_EInp_GPU) * cp.fft.fft2(KIMG_CSZ_GPU)
            ).real[W0: -W0, W1: -W1]

        if FORCE_OUTPUT_C_CONTIGUOUS:
            if not PixA_Out_GPU.flags['C_CONTIGUOUS']:
                PixA_Out_GPU = cp.ascontiguousarray(PixA_Out_GPU)
        return PixA_Out_GPU



import time
import warnings
import numpy as np
import os.path as pa
from astropy.io import fits
import scipy.ndimage as ndimage
from astropy.table import Column
from sfft.AutoSparsePrep import Auto_SparsePrep
from sfft.utils.SFFTSolutionReader import Realize_FluxScaling
from sfft.sfftcore.SFFTSubtract import GeneralSFFTSubtract
from sfft.sfftcore.SFFTConfigure import SingleSFFTConfigure
# version: Mar 18, 2023

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.4"

class Easy_SparsePacket:
    @staticmethod
    def ESP(FITS_REF, FITS_SCI, FITS_DIFF=None, FITS_Solution=None, ForceConv='AUTO', GKerHW=None, \
        KerHWRatio=2.0, KerHWLimit=(2, 20), KerPolyOrder=2, BGPolyOrder=0, ConstPhotRatio=True, \
        MaskSatContam=False, GAIN_KEY='GAIN', SATUR_KEY='ESATUR', BACK_TYPE='MANUAL', BACK_VALUE=0.0, \
        BACK_SIZE=64, BACK_FILTERSIZE=3, DETECT_THRESH=2.0, ANALYSIS_THRESH=2.0, DETECT_MINAREA=5, \
        DETECT_MAXAREA=0, DEBLEND_MINCONT=0.005, BACKPHOTO_TYPE='LOCAL', ONLY_FLAGS=[0], BoundarySIZE=30, \
        XY_PriorSelect=None, Hough_MINFR=0.1, Hough_PeakClip=0.7, BeltHW=0.2, PointSource_MINELLIP=0.3, \
        MatchTol=None, MatchTolFactor=3.0, COARSE_VAR_REJECTION=True, CVREJ_MAGD_THRESH=0.12, \
        ELABO_VAR_REJECTION=True, EVREJ_RATIO_THREH=5.0, EVREJ_SAFE_MAGDEV=0.04, StarExt_iter=4, \
        XY_PriorBan=None, PostAnomalyCheck=False, PAC_RATIO_THRESH=5.0, BACKEND_4SUBTRACT='Cupy', \
        CUDA_DEVICE_4SUBTRACT='0', NUM_CPU_THREADS_4SUBTRACT=8, VERBOSE_LEVEL=2):
        
        """
        # NOTE: This function is to Perform Sparse-Flavor SFFT for a single task:
        #       Cupy backend: do preprocessing on one CPU thread, and perform subtraction on one GPU device.
        #       Numpy backend: do preprocessing on one CPU thread, and perform subtraction with pyFFTW and Numba 
        #                      using multiple threads (-NUM_CPU_THREADS_4SUBTRACT).

        * Parameters for Sparse-Flavor SFFT [single task]

        # ----------------------------- Computing Enviornment --------------------------------- #

        -BACKEND_4SUBTRACT ['Cupy']     # 'Cupy' or 'Numpy'. 
                                        # Cupy backend require GPU(s) that is capable of performing double-precision calculations,
                                        # while Numpy backend is a pure CPU-based backend for sfft subtraction.
                                        # NOTE: 'Pycuda' backend is no longer supported since sfft v1.4.0.
        
        -CUDA_DEVICE_4SUBTRACT ['0']    # it specifies certain GPU device (index) to conduct the subtraction task.
                                        # the GPU devices are usually numbered 0 to N-1 (you may use command nvidia-smi to check).
                                        # this argument becomes trivial for Numpy backend.

        -NUM_CPU_THREADS_4SUBTRACT [8]  # it specifies the number of CPU threads used for sfft subtraction in Numpy backend.
                                        # SFFT in Numpy backend has been implemented with pyFFTW and numba, 
                                        # that allow for parallel computing on CPUs. Of course, the Numpy 
                                        # backend is generally much slower than GPU backends.
        
        # ----------------------------- Preprocessing with Source Selection for Image-Masking --------------------------------- #

        # > Configurations for SExtractor

        -GAIN_KEY ['GAIN']             # SExtractor Parameter GAIN_KEY
                                       # i.e., keyword of GAIN in FITS header (of reference & science)

        -SATUR_KEY ['ESATUR']          # SExtractor Parameter SATUR_KEY
                                       # i.e., keyword of effective saturation in FITS header (of reference & science)
                                       # Remarks: one may think 'SATURATE' is a more common keyword name for saturation level.
                                       #          However, note that Sparse-Flavor SFFT requires sky-subtracted images as inputs, 
                                       #          we need to use the 'effective' saturation level after the sky-subtraction.
                                       #          e.g., set ESATURA = SATURATE - (SKY + 10*SKYSIG)

        -BACK_TYPE ['MANUAL']          # SExtractor Parameter BACK_TYPE = [AUTO or MANUAL].
                                       # As Sparse-Flavor-SFFT requires the input images being sky-subtracted,
                                       # the default setting uses zero-background (i.e., BACK_TYPE='MANUAL' & BACK_VALUE=0.0).
                                       # However, one may also use BACK_TYPE='AUTO' for some specific cases
                                       # E.g., point sources located at the outskirts of a bright galaxy will have a relatively
                                       #       high local background. Even if the image has been sky-subtracted properly, 
                                       #       SExtractor can only detect it when you set BACK_TYPE='AUTO'.
         
        -BACK_VALUE [0.0]              # SExtractor Parameter BACK_VALUE (only work for BACK_TYPE='MANUAL')

        -BACK_SIZE [64]                # SExtractor Parameter BACK_SIZE

        -BACK_FILTERSIZE [3]           # SExtractor Parameter BACK_FILTERSIZE

        -DETECT_THRESH [2.0]           # SExtractor Parameter DETECT_THRESH
                                       # NOTE: One may notice that the default DETECT_THRESH in SExtractor is 1.5,
                                       #       Using a 'cold' detection threshold here is to speed up SExtractor.
                                       #       Although DETECT_THRESH = 2.0 means we will miss the faint-end sources with 
                                       #       SNR < 9 (approximately), the cost is generally acceptable for source selection.

        -ANALYSIS_THRESH [2.0]         # SExtractor Parameter ANALYSIS_THRESH
                                       # NOTE: By default, let -ANALYSIS_THRESH = -DETECT_THRESH

        -DETECT_MINAREA [5]            # SExtractor Parameter DETECT_MINAREA
        
        -DETECT_MAXAREA [0]            # SExtractor Parameter DETECT_MAXAREA

        -DEBLEND_MINCONT [0.005]       # SExtractor Parameter DEBLEND_MINCONT (typically, 0.001 - 0.005)

        -BACKPHOTO_TYPE ['LOCAL']      # SExtractor Parameter BACKPHOTO_TYPE

        -ONLY_FLAGS [0]                # Restrict SExtractor Output Photometry Catalog by Source FLAGS
                                       # Common FLAGS (Here None means no restrictions on FLAGS):
                                       # 1: aperture photometry is likely to be biased by neighboring sources 
                                       #    or by more than 10% of bad pixels in any aperture
                                       # 2: the object has been deblended
                                       # 4: at least one object pixel is saturated

        -BoundarySIZE [30]             # Restrict SExtractor Output Photometry Catalog by Dropping Sources at Boundary 
                                       # NOTE: This would help to avoid selecting sources too close to image boundary. 

        # > Configurations for Source-Selction
        #
        # Remarks on two modes 
        #   [HOUGH-AUTO] MODE: Source-Selection based on SExtractor & Hough Transformation, following Hu, et al. (2022).
        #   [SEMI-AUTO] MODE:  Source-Selection directly make use of a prior-selected source list.
        #

        -XY_PriorSelect [None]         # a Numpy array of pixels coordinates, with shape (N, 2) (e.g., [[x0, y0], [x1, y1], ...])
                                       # this allows sfft to us a prior source selection to solve the subtraction.
                                       # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [SEMI-AUTO] MODE

        -Hough_MINFR [0.1]             # The lower bound of FLUX_RATIO for line feature detection using Hough transformation.
                                       # Setting a proper lower bound can avoid to detect some line features by chance,
                                       # which are not contributed from point sources but resides in the small-FLUX_RATIO region.
                                       # NOTE: Recommended values of Hough_MINFR: 0.1 ~ 1.0
                                       # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE

        -Hough_PeakClip [0.7]          # It determines the lower bound of the sensitivity of the line feature detection.
                                       # NOTE: When the point-source-belt is not very pronounced (e.g., in galaxy dominated fields),
                                       #       one may consider to reduce the parameter from default 0.7 to, says, ~ 0.4.
                                       # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE
    
        -BeltHW [0.2]                  # The half-width of point-source-belt detected by Hough Transformation.
                                       # Remarks: if you want to tune this parameter, it is helpful to draw 
                                       #          a figure of MAG_AUTO against FLUX_RADIUS.
                                       # NOTE: IT WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE :::: determine point-source-belt

        -PointSource_MINELLIP [0.3]    # An additiona Restriction on ELLIPTICITY (ELLIPTICITY < PointSource_MINELLIP) 
                                       # for point sources.
                                       # NOTE: IT WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE :::: determine point-sources

        -MatchTol [None]               # Given separation tolerance (pix) for source matching 
                                       # NOTE: IT WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE :::: Cross-Match between REF & SCI
                                       # NOTE: IT WORKS FOR Auto-Sparse-Prep [SEMI-AUTO] MODE :::: Cross-Match between REF & SCI AND
                                       #       Cross-Match between the prior selection and mean image coordinates of matched REF-SCI sources.

        -MatchTolFactor [3.0]          # The separation tolerance (pix) for source matching calculated by FWHM,
                                       # MatchTol = np.sqrt((FWHM_REF/MatchTolFactor)**2 + (FWHM_SCI/MatchTolFactor)**2)
                                       # @ Given precise WCS, one can use a high MatchTolFactor ~3.0
                                       # @ For very sparse fields where WCS can be inaccurate, 
                                       #   one can loosen the tolerance with a low MatchTolFactor ~1.0
                                       # NOTE: IT WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE :::: Cross-Match between REF & SCI
                                       # NOTE: IT WORKS FOR Auto-Sparse-Prep [SEMI-AUTO] MODE :::: Cross-Match between REF & SCI AND
                                       #       Cross-Match between the prior selection and mean image coordinates of matched REF-SCI sources.

        -COARSE_VAR_REJECTION [True]   # Activate Coarse Variable Rejection (CVREJ) or not.
                                       # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE 

        -CVREJ_MAGD_THRESH [0.12]      # CVREJ kicks out the obvious variables by the difference of 
                                       # instrument magnitudes (MAG_AUTO) measured on reference and science, i.e., 
                                       # MAG_AUTO_SCI - MAG_AUTO_REF. A source with difference highly deviated from
                                       # the median level of the field stars will be rejected from the source selection.
                                       # -CVREJ_MAGD_THRESH is the deviation magnitude threshold. 
                                       # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE 
        
        -ELABO_VAR_REJECTION [True]    # WARNING: EVREJ REQUIRES CORRECT GAIN IN FITS HEADER!
                                       # Activate Elaborate Variable Rejection (EVREJ) or not.
                                       # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE 

        -EVREJ_RATIO_THREH [5.0]       # EVREJ kicks out the variables by the photometric uncertainties
                                       # (SExtractor FLUXERR_AUTO) measured on reference and science.
                                       # The flux change of a stationary object from reference to science 
                                       # is a predictable probability ditribution. EVREJ will reject the 
                                       # -EVREJ_RATIO_THREH * SIGMA outliers of the probability distribution.
                                       # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE 

        -EVREJ_SAFE_MAGDEV [0.04]      # A protection to avoid EVREJ overkilling. EVREJ would not reject
                                       # a source if the difference of instrument magnitudes is deviated
                                       # from the median level within -EVREJ_SAFE_MAGDEV magnitude.
                                       # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE 
                                       #
                                       # @ Why does EVREJ overkill?
                                       # i) The calculated flux change is entangled with the error of 
                                       #    the constant photometric scaling used by EVREJ.
                                       # ii) SExtractor FLUXERR can be in accurate, especially at the bright end.
                                       # see more details in sfft.Auto_SparsePrep.HoughAutoMask

        -StarExt_iter [4]              # Our image mask is determined by the SExtractor check image SEGMENTATION 
                                       # of the selected sources. note that some pixels (e.g., outskirt region of a galaxy) 
                                       # harbouring signal may be not SExtractor-detectable due to the nature of 
                                       # the thresholding-based detection method (see NoiseChisel paper for more details). 
                                       # we want to include the missing light at outskirt region to contribute to
                                       # parameter-solving process, then a simple mask dilation is introduced.
                                       # -StarExt_iter means the iteration times of the dilation process.
                                       # NOTE: IT WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE :::: mask dilation
                                       # NOTE: IT WORKS FOR Auto-Sparse-Prep [SEMI-AUTO] MODE :::: mask dilation

        -XY_PriorBan [None]            # a Numpy array of pixels coordinates, with shape (N, 2) (e.g., [[x0, y0], [x1, y1], ...])
                                       # this allows us to feed the prior knowledge about the varibility cross the field.
                                       # if you already get a list of variables (transients) and would like SFFT not to select
                                       # them to solve the subtraction, you can tell SFFT their coordinates through -XY_PriorBan.
                                       # NOTE: IT WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE :::: mask refinement
                                       # NOTE: IT WORKS FOR Auto-Sparse-Prep [SEMI-AUTO] MODE :::: mask refinement

        # ----------------------------- SFFT Subtraction --------------------------------- #

        -ForceConv ['AUTO']         # it determines which image will be convolved, can be 'REF', 'SCI' and 'AUTO'.
                                    # -ForceConv = 'AUTO' means SFFT will determine the convolution direction according to 
                                    # FWHM_SCI and FWHM_REF: the image with better seeing will be convolved to avoid deconvolution.

        -GKerHW [None]              # the given kernel half-width, None means the kernel size will be 
                                    # automatically determined by -KerHWRatio (to be seeing-related). 

        -KerHWRatio [2.0]           # the ratio between FWHM and the kernel half-width
                                    # KerHW = int(KerHWRatio * Max(FWHM_REF, FWHM_SCI))

        -KerHWLimit [(2, 20)]       # the lower & upper bounds for kernel half-width 
                                    # KerHW is updated as np.clip(KerHW, KerHWLimit[0], KerHWLimit[1]) 
                                    # Remarks: this is useful for a survey since it can constrain the peak GPU memory usage.

        -KerPolyOrder [2]           # Polynomial degree of kernel spatial variation, can be [0,1,2,3].

        -BGPolyOrder [0]            # Polynomial degree of differential-background spatial variation, can be [0,1,2,3].
                                    # NOTE: this argument is trivial for Sparse-Flavor SFFT as input images have been sky subtracted.
                                    #       simply use a flat differential background (it will very close to 0 in the sfft solution). 

        -ConstPhotRatio [True]      # Constant photometric ratio between images? can be True or False
                                    # ConstPhotRatio = True: the sum of convolution kernel is restricted 
                                    #   to be a constant across the field. 
                                    # ConstPhotRatio = False: the flux scaling between images is modeled
                                    #   by a polynomial with degree -KerPolyOrder.

        -MaskSatContam [False]      # Mask saturation-contaminated regions on difference image ? can be True or False
                                    # NOTE the pixels enclosed in the regions are replaced by NaN.

        # ----------------------------- Post Subtraction --------------------------------- #

        -PostAnomalyCheck [False]   # WARNING: PAC REQUIRES CORRECT GAIN IN FITS HEADER!
                                    # Post Anomaly Check (PAC) is to find the missing variables in the source selection 
                                    # on the difference image. Like EVREJ, it depends on the photometric uncertainties 
                                    # given by SExtractor FLUXERR_AUTO of reference and science image.
                                    # As image subtraction can get a more accurate photometric scaling between reference 
                                    # and science than that in EVREJ, PA is a even more 'elaborate' way to identify variables.
                                    # NOTE: one can make use of the Post-Anomalies as the Prior-Ban-Sources 
                                    #       to refine the subtraction (run second-time subtraction).
        
        -PAC_RATIO_THRESH [5.0]     # PAC will identify the -PAC_RATIO_THRESH * SIGMA outliers of the 
                                    # probability distribution as missing variables.
        
        # ----------------------------- Input & Output --------------------------------- #

        -FITS_REF []                # File path of input reference image

        -FITS_SCI []                # File path of input science image

        -FITS_DIFF [None]           # File path of output difference image

        -FITS_Solution [None]       # File path of the solution of the linear system 
                                    # it is an array of (..., a_ijab, ... b_pq, ...)

        # ----------------------------- Miscellaneous --------------------------------- #
        
        -VERBOSE_LEVEL [2]          # The level of verbosity, can be [0, 1, 2]
                                    # 0/1/2: QUIET/NORMAL/FULL mode

        # Important Notice:
        #
        # a): if reference is convolved in SFFT, then DIFF = SCI - Convolved_REF.
        #     [difference image is expected to have PSF & flux zero-point consistent with science image]
        #     e.g., -ForceConv='REF' or -ForceConv='AUTO' when reference has better seeing.
        #
        # b): if science is convolved in SFFT, then DIFF = Convolved_SCI - REF
        #     [difference image is expected to have PSF & flux zero-point consistent with reference image]
        #     e.g., -ForceConv='SCI' or -ForceConv='AUTO' when science has better seeing.
        #
        # Remarks: this convention is to guarantee that a transient emerge on science image 
        #          always shows itself as a positive signal on the difference images.

        """
        
        # * Perform Auto Sparse-Prep
        if VERBOSE_LEVEL in [2]:
            warnings.warn('\nMeLOn REMINDER: Input images for sparse-flavor sfft should be SKY-SUBTRACTED!')
        
        _ASP = Auto_SparsePrep(FITS_REF=FITS_REF, FITS_SCI=FITS_SCI, GAIN_KEY=GAIN_KEY, SATUR_KEY=SATUR_KEY, \
            BACK_TYPE=BACK_TYPE, BACK_VALUE=BACK_VALUE, BACK_SIZE=BACK_SIZE, BACK_FILTERSIZE=BACK_FILTERSIZE, \
            DETECT_THRESH=DETECT_THRESH, ANALYSIS_THRESH=ANALYSIS_THRESH, DETECT_MINAREA=DETECT_MINAREA, \
            DETECT_MAXAREA=DETECT_MAXAREA, DEBLEND_MINCONT=DEBLEND_MINCONT, BACKPHOTO_TYPE=BACKPHOTO_TYPE, \
            ONLY_FLAGS=ONLY_FLAGS, BoundarySIZE=BoundarySIZE, VERBOSE_LEVEL=VERBOSE_LEVEL)

        if XY_PriorSelect is None:
            IMAGE_MASK_METHOD = 'HOUGH-AUTO'
            if VERBOSE_LEVEL in [0, 1, 2]:
                print('MeLOn CheckPoint: TRIGGER Sparse-Flavor Auto Preprocessing [%s] MODE!' %IMAGE_MASK_METHOD)

            SFFTPrepDict = _ASP.HoughAutoMask(Hough_MINFR=Hough_MINFR, Hough_PeakClip=Hough_PeakClip, \
                BeltHW=BeltHW, PointSource_MINELLIP=PointSource_MINELLIP, MatchTol=MatchTol, MatchTolFactor=MatchTolFactor, \
                COARSE_VAR_REJECTION=COARSE_VAR_REJECTION, CVREJ_MAGD_THRESH=CVREJ_MAGD_THRESH, \
                ELABO_VAR_REJECTION=ELABO_VAR_REJECTION, EVREJ_RATIO_THREH=EVREJ_RATIO_THREH, \
                EVREJ_SAFE_MAGDEV=EVREJ_SAFE_MAGDEV, StarExt_iter=StarExt_iter, XY_PriorBan=XY_PriorBan)
        else:
            IMAGE_MASK_METHOD = 'SEMI-AUTO'
            if VERBOSE_LEVEL in [0, 1, 2]:
                print('MeLOn CheckPoint: TRIGGER Sparse-Flavor Auto Preprocessing [%s] MODE!' %IMAGE_MASK_METHOD)
            
            SFFTPrepDict = _ASP.SemiAutoMask(XY_PriorSelect=XY_PriorSelect, MatchTol=MatchTol, \
                MatchTolFactor=MatchTolFactor, StarExt_iter=StarExt_iter, XY_PriorBan=XY_PriorBan)

        # * Determine ConvdSide & KerHW
        FWHM_REF = SFFTPrepDict['FWHM_REF']
        FWHM_SCI = SFFTPrepDict['FWHM_SCI']

        assert ForceConv in ['AUTO', 'REF', 'SCI']
        if ForceConv == 'AUTO':
            if FWHM_SCI >= FWHM_REF: ConvdSide = 'REF'
            else: ConvdSide = 'SCI'
        else: ConvdSide = ForceConv

        if GKerHW is None:
            FWHM_La = np.max([FWHM_REF, FWHM_SCI])
            KerHW = int(np.clip(KerHWRatio * FWHM_La, KerHWLimit[0], KerHWLimit[1]))
        else: KerHW = GKerHW

        # * Choose GPU device for Cupy backend
        if BACKEND_4SUBTRACT == 'Cupy':
            import cupy as cp
            device = cp.cuda.Device(int(CUDA_DEVICE_4SUBTRACT))
            device.use()
        
        # * Compile Functions in SFFT Subtraction
        PixA_REF = SFFTPrepDict['PixA_REF']
        PixA_SCI = SFFTPrepDict['PixA_SCI']

        if VERBOSE_LEVEL in [0, 1, 2]:
            print('MeLOn CheckPoint: TRIGGER Function Compilations of SFFT-SUBTRACTION!')

        Tcomp_start = time.time()
        SFFTConfig = SingleSFFTConfigure.SSC(NX=PixA_REF.shape[0], NY=PixA_REF.shape[1], KerHW=KerHW, \
            KerPolyOrder=KerPolyOrder, BGPolyOrder=BGPolyOrder, ConstPhotRatio=ConstPhotRatio, \
            BACKEND_4SUBTRACT=BACKEND_4SUBTRACT, NUM_CPU_THREADS_4SUBTRACT=NUM_CPU_THREADS_4SUBTRACT, \
            VERBOSE_LEVEL=VERBOSE_LEVEL)
        
        if VERBOSE_LEVEL in [1, 2]:
            _message = 'Function Compilations of SFFT-SUBTRACTION '
            _message += 'TAKES [%.3f s]!' %(time.time() - Tcomp_start)
            print('\nMeLOn Report: %s' %_message)
        
        # * Perform SFFT Subtraction
        SatMask_REF = SFFTPrepDict['REF-SAT-Mask']
        SatMask_SCI = SFFTPrepDict['SCI-SAT-Mask']
        NaNmask_U = SFFTPrepDict['Union-NaN-Mask']
        PixA_mREF = SFFTPrepDict['PixA_mREF']
        PixA_mSCI = SFFTPrepDict['PixA_mSCI']

        if ConvdSide == 'REF':
            PixA_mI, PixA_mJ = PixA_mREF, PixA_mSCI
            if NaNmask_U is not None:
                PixA_I, PixA_J = PixA_REF.copy(), PixA_SCI.copy()
                PixA_I[NaNmask_U] = PixA_mI[NaNmask_U]
                PixA_J[NaNmask_U] = PixA_mJ[NaNmask_U]
            else: PixA_I, PixA_J = PixA_REF, PixA_SCI
            if MaskSatContam: 
                ContamMask_I = SatMask_REF
                ContamMask_J = SatMask_SCI
            else: ContamMask_I = None

        if ConvdSide == 'SCI':
            PixA_mI, PixA_mJ = PixA_mSCI, PixA_mREF
            if NaNmask_U is not None:
                PixA_I, PixA_J = PixA_SCI.copy(), PixA_REF.copy()
                PixA_I[NaNmask_U] = PixA_mI[NaNmask_U]
                PixA_J[NaNmask_U] = PixA_mJ[NaNmask_U]
            else: PixA_I, PixA_J = PixA_SCI, PixA_REF
            if MaskSatContam: 
                ContamMask_I = SatMask_SCI
                ContamMask_J = SatMask_REF
            else: ContamMask_I = None

        if VERBOSE_LEVEL in [0, 1, 2]:
            print('MeLOn CheckPoint: TRIGGER SFFT-SUBTRACTION!')

        Tsub_start = time.time()
        _tmp = GeneralSFFTSubtract.GSS(PixA_I=PixA_I, PixA_J=PixA_J, PixA_mI=PixA_mI, PixA_mJ=PixA_mJ, \
            SFFTConfig=SFFTConfig, ContamMask_I=ContamMask_I, BACKEND_4SUBTRACT=BACKEND_4SUBTRACT, \
            NUM_CPU_THREADS_4SUBTRACT=NUM_CPU_THREADS_4SUBTRACT, VERBOSE_LEVEL=VERBOSE_LEVEL)
        
        Solution, PixA_DIFF, ContamMask_CI = _tmp
        if VERBOSE_LEVEL in [1, 2]:
            _message = 'SFFT-SUBTRACTION TAKES [%.3f s]!' %(time.time() - Tsub_start)
            print('\nMeLOn Report: %s' %_message)
        
        # ** adjust the sign of difference image 
        #    NOTE: Our convention is to make the transients on SCI always show themselves as positive signal on DIFF
        #    (a) when REF is convolved, DIFF = SCI - Conv(REF)
        #    (b) when SCI is convolved, DIFF = Conv(SCI) - REF

        if ConvdSide == 'SCI':
            PixA_DIFF = -PixA_DIFF

        if VERBOSE_LEVEL in [1, 2]:
            if ConvdSide == 'REF':
                _message = 'Reference Image is Convolved in SFFT-SUBTRACTION [DIFF = SCI - Conv(REF)]!'
                print('MeLOn CheckPoint: %s' %_message)

            if ConvdSide == 'SCI':
                _message = 'Science Image is Convolved in SFFT-SUBTRACTION [DIFF = Conv(SCI) - REF]!'
                print('MeLOn CheckPoint: %s' %_message)

        # ** estimate the flux scaling through the convolution of image subtraction
        #    NOTE: if photometric ratio is not constant (ConstPhotRatio=False), we measure of a grid
        #          of coordinates to estimate the flux scaling and its fluctuation (polynomial form).
        #    NOTE: here we set the tile-size of the grid about 64 x 64 pix, but meanwhile, 
        #          we also require the tiles should no less than 6 along each axis.

        N0, N1 = SFFTConfig[0]['N0'], SFFTConfig[0]['N1']
        L0, L1 = SFFTConfig[0]['L0'], SFFTConfig[0]['L1']
        DK, Fpq = SFFTConfig[0]['DK'], SFFTConfig[0]['Fpq']

        if ConstPhotRatio:
            SFFT_FSCAL_NSAMP = 1
            XY_q = np.array([[N0/2.0, N1/2.0]]) + 0.5
            RFS = Realize_FluxScaling(XY_q=XY_q)
            SFFT_FSCAL_q = RFS.FromArray(Solution=Solution, N0=N0, N1=N1, L0=L0, L1=L1, DK=DK, Fpq=Fpq)
            SFFT_FSCAL_MEAN, SFFT_FSCAL_SIG = SFFT_FSCAL_q[0], 0.0
        
        if not ConstPhotRatio:
            TILESIZE_X, TILESIZE_Y = 64, 64    # FIXME emperical/arbitrary values
            NTILE_X = np.max([round(N0/TILESIZE_X), 6])
            NTILE_Y = np.max([round(N1/TILESIZE_Y), 6])
            GX = np.linspace(0.5, N0+0.5, NTILE_X+1)
            GY = np.linspace(0.5, N1+0.5, NTILE_Y+1)
            YY, XX = np.meshgrid(GY, GX)
            XY_q = np.array([XX.flatten(), YY.flatten()]).T
            SFFT_FSCAL_NSAMP = XY_q.shape[0]
            
            RFS = Realize_FluxScaling(XY_q=XY_q)
            SFFT_FSCAL_q = RFS.FromArray(Solution=Solution, N0=N0, N1=N1, L0=L0, L1=L1, DK=DK, Fpq=Fpq)
            SFFT_FSCAL_MEAN, SFFT_FSCAL_SIG = np.mean(SFFT_FSCAL_q), np.std(SFFT_FSCAL_q)

        PHOT_FSCAL = 10**(SFFTPrepDict['MAG_OFFSET']/-2.5)  # FLUX_SCI/FLUX_REF
        if ConvdSide == 'SCI': PHOT_FSCAL = 1.0/PHOT_FSCAL

        if VERBOSE_LEVEL in [1, 2]:
            _message = 'The Flux Scaling through the Convolution of SFFT-SUBTRACTION '
            _message += '[%.6f +/- %.6f] from [%d] positions!\n' %(SFFT_FSCAL_MEAN, SFFT_FSCAL_SIG, SFFT_FSCAL_NSAMP)
            _message += 'P.S. The approximated Flux Scaling from Photometry [%.6f].' %PHOT_FSCAL
            print('MeLOn CheckPoint: %s' %_message)
        
        # ** run post anomaly check (PAC)
        if PostAnomalyCheck:

            """
            # Remarks on the Post Anomaly Check (PAC)
            # [1] One may notice that Elaborate Variable Rejection (EVREJ) also use SExtractor 
            #     photometric errors to identify variables. However, PAC should be more accurate than EVREJ.
            #     NOTE: Bright transients (compared with its host) can be also identified by PAC.
            #
            # [2] distribution mean: image subtraction is expected to have better accuracy on the determination 
            #     of the photometric scaling, compared with the constant MAG_OFFSET in EVREJ. As a result,
            #     the flux change counted on the difference image is likely more 'real' than that in EVREJ.
            #
            # [3] distribution variance: SExtractor FLUXERR_AUTO can be inaccurate, and the convolution effect
            #     on the photometric errors has been simplified as multiplying an average flux scaling on an image.
            #
            """

            if VERBOSE_LEVEL in [2]:
                warnings.warn('\nMeLOn REMINDER: Post-Anomaly Check requires CORRECT GAIN in FITS HEADER!')

            AstSEx_SS = SFFTPrepDict['SExCatalog-SubSource']
            SFFTLmap = SFFTPrepDict['SFFT-LabelMap']
            
            # ** Only consider the Non-Prior-Banned SubSources
            if 'MASK_PriorBan' in AstSEx_SS.colnames:
                nPBMASK_SS = ~np.array(AstSEx_SS['MASK_PriorBan'])
                AstSEx_vSS = AstSEx_SS[nPBMASK_SS]
            else: AstSEx_vSS = AstSEx_SS
            
            # ** Estimate expected variance of Non-Prior-Banned SubSources on the difference
            FLUXERR_vSSr = np.array(AstSEx_vSS['FLUXERR_AUTO_REF'])
            FLUXERR_vSSs = np.array(AstSEx_vSS['FLUXERR_AUTO_SCI'])

            if ConvdSide == 'REF':
                sFLUXERR_vSSr = FLUXERR_vSSr * SFFT_FSCAL_MEAN
                ExpDVAR_vSS = sFLUXERR_vSSr**2 + FLUXERR_vSSs**2

            if ConvdSide == 'SCI':
                sFLUXERR_vSSs = FLUXERR_vSSs * SFFT_FSCAL_MEAN
                ExpDVAR_vSS = FLUXERR_vSSr**2 + sFLUXERR_vSSs**2

            # ** Measure the ratios of Non-Prior-Banned SubSources on the difference for PAC
            SEGL_vSS = np.array(AstSEx_vSS['SEGLABEL']).astype(int)
            DFSUM_vSS = ndimage.labeled_comprehension(PixA_DIFF, SFFTLmap, SEGL_vSS, np.sum, float, 0.0)
            RATIO_vSS = DFSUM_vSS / np.clip(np.sqrt(ExpDVAR_vSS), a_min=1e-8, a_max=None)
            PAMASK_vSS = np.abs(RATIO_vSS) > PAC_RATIO_THRESH

            if VERBOSE_LEVEL in [1, 2]:
                _message = 'Identified [%d] PostAnomaly SubSources [> %.2f sigma] ' %(np.sum(PAMASK_vSS), PAC_RATIO_THRESH)
                _message += 'out of [%d] Non-Prior-Banned SubSources!\n' %(len(AstSEx_vSS))
                _message += 'P.S. there are [%d] Prior-Banned SubSources!' %(len(AstSEx_SS) - len(AstSEx_vSS))
                print('\nMeLOn CheckPoint: %s' %_message)
            
            # ** Record the results (decorate AstSEx_SS in SFFTPrepDict)
            if 'MASK_PriorBan' in AstSEx_SS.colnames:
                ExpDVAR_SS = np.nan * np.ones(len(AstSEx_SS))       # NOTE Prior-Ban is trivial NaN
                DFSUM_SS = np.nan * np.ones(len(AstSEx_SS))         # NOTE Prior-Ban is trivial NaN
                RATIO_SS = np.nan * np.ones(len(AstSEx_SS))         # NOTE Prior-Ban is trivial NaN
                PAMASK_SS = np.zeros(len(AstSEx_SS)).astype(bool)   # NOTE Prior-Ban is trivial False
                
                ExpDVAR_SS[nPBMASK_SS] = ExpDVAR_vSS
                DFSUM_SS[nPBMASK_SS] = DFSUM_vSS
                RATIO_SS[nPBMASK_SS] = RATIO_vSS
                PAMASK_SS[nPBMASK_SS] = PAMASK_vSS
            else: 
                ExpDVAR_SS = ExpDVAR_vSS
                DFSUM_SS = DFSUM_vSS
                RATIO_SS = RATIO_vSS
                PAMASK_SS = PAMASK_vSS

            AstSEx_SS.add_column(Column(ExpDVAR_SS, name='ExpDVAR_PostAnomaly'))
            AstSEx_SS.add_column(Column(DFSUM_SS, name='DFSUM_PostAnomaly'))
            AstSEx_SS.add_column(Column(RATIO_SS, name='RATIO_PostAnomaly'))
            AstSEx_SS.add_column(Column(PAMASK_SS, name='MASK_PostAnomaly'))
        
        # ** final tweaks on the difference image
        if NaNmask_U is not None:
            PixA_DIFF[NaNmask_U] = np.nan   # Mask Union-NaN regions
        
        if MaskSatContam:
            ContamMask_DIFF = np.logical_or(ContamMask_CI, ContamMask_J)
            PixA_DIFF[ContamMask_DIFF] = np.nan   # Mask Saturation-Contaminated regions
        
        # * Save difference image
        if FITS_DIFF is not None:

            """
            # Remarks on header of difference image
            # [1] In general, the FITS header of difference image would mostly be inherited from science image.
            #
            # [2] when science image is convolved, we turn to set GAIN_DIFF = GAIN_SCI / SFFT_FSCAL_MEAN (unit: e-/ADU)
            #     so that one can correctly convert ADU to e- for a transient appears on science image.
            #     WARNING: for a transient appears on reference image, GAIN_DIFF = GAIN_SCI is correct.
            #              we should keep in mind that GAIN_DIFF is not an absolute instrumental value.
            #     WARNING: one can only estimate the Poission noise from the science transient via GIAN_DIFF.
            #              the Poission noise from the background source (host galaxy) can enhance the 
            #              background noise at the position of the transient. photometry softwares which 
            #              only take background noise and transient Possion noise into account would tend 
            #              to overestimate the SNR of the transient.
            # 
            # [3] when science image is convolved, we turn to set SATURA_DIFF = SATURA_SCI * SFFT_FSCAL_MEAN         
            #     WARNING: it is still not a good idea to use SATURA_DIFF to identify the SATURATION regions
            #              on difference image. more appropriate way is masking the saturation contaminated 
            #              regions on difference image (MaskSatContam=True), or alternatively, leave them  
            #              alone and using AI stamp classifier to reject saturation-related bogus.
            #
            """

            _hdl = fits.open(FITS_SCI)
            _hdl[0].data[:, :] = PixA_DIFF.T

            _hdl[0].header['NAME_REF'] = (pa.basename(FITS_REF), 'MeLOn: SFFT')
            _hdl[0].header['NAME_SCI'] = (pa.basename(FITS_SCI), 'MeLOn: SFFT')
            _hdl[0].header['FWHM_REF'] = (FWHM_REF, 'MeLOn: SFFT')
            _hdl[0].header['FWHM_SCI'] = (FWHM_SCI, 'MeLOn: SFFT')
            _hdl[0].header['KERORDER'] = (KerPolyOrder, 'MeLOn: SFFT')
            _hdl[0].header['BGORDER'] = (BGPolyOrder, 'MeLOn: SFFT')
            _hdl[0].header['CPHOTR'] = (str(ConstPhotRatio), 'MeLOn: SFFT')
            _hdl[0].header['KERHW'] = (KerHW, 'MeLOn: SFFT')
            _hdl[0].header['CONVD'] = (ConvdSide, 'MeLOn: SFFT')

            if ConvdSide == 'SCI':
                GAIN_SCI = _hdl[0].header[GAIN_KEY]
                SATUR_SCI = _hdl[0].header[SATUR_KEY]
                GAIN_DIFF = GAIN_SCI / SFFT_FSCAL_MEAN
                SATUR_DIFF = SATUR_SCI * SFFT_FSCAL_MEAN

                _hdl[0].header[GAIN_KEY] = (GAIN_DIFF, 'MeLOn: SFFT')
                _hdl[0].header[SATUR_KEY] = (SATUR_DIFF, 'MeLOn: SFFT')

            _hdl.writeto(FITS_DIFF, overwrite=True)
            _hdl.close()
        
        # * Save solution array
        if FITS_Solution is not None:
            phdu = fits.PrimaryHDU()
            phdu.header['N0'] = (SFFTConfig[0]['N0'], 'MeLOn: SFFT')
            phdu.header['N1'] = (SFFTConfig[0]['N1'], 'MeLOn: SFFT')
            phdu.header['DK'] = (SFFTConfig[0]['DK'], 'MeLOn: SFFT')
            phdu.header['DB'] = (SFFTConfig[0]['DB'], 'MeLOn: SFFT')
            phdu.header['L0'] = (SFFTConfig[0]['L0'], 'MeLOn: SFFT')
            phdu.header['L1'] = (SFFTConfig[0]['L1'], 'MeLOn: SFFT')
            
            phdu.header['FIJ'] = (SFFTConfig[0]['Fij'], 'MeLOn: SFFT')
            phdu.header['FAB'] = (SFFTConfig[0]['Fab'], 'MeLOn: SFFT')
            phdu.header['FPQ'] = (SFFTConfig[0]['Fpq'], 'MeLOn: SFFT')
            phdu.header['FIJAB'] = (SFFTConfig[0]['Fijab'], 'MeLOn: SFFT')

            PixA_Solution = Solution.reshape((-1, 1))
            phdu.data = PixA_Solution.T
            fits.HDUList([phdu]).writeto(FITS_Solution, overwrite=True)

        return PixA_DIFF, SFFTPrepDict, Solution, SFFT_FSCAL_MEAN, SFFT_FSCAL_SIG



import numpy as np
import importlib.util as iu
from sfft.utils.PureCupyFFTKits import PureCupy_FFTKits
if iu.find_spec('cupy') is not None:
    import cupy as cp

__last_update__ = "2024-09-22"
__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.6.1"

#########################################################################
#||                                                                   ||#
#||                           Acknowledgment                          ||#
#||                                                                   ||#
#########################################################################
#||                                                                   ||#
#||   This code was developed during the NASA GPU Hackathon 2024      ||#
#||   (https://www.nas.nasa.gov/hackathon/#home) as part of the       ||#
#||   pipeline optimization efforts for the Roman Supernova PIT       ||#
#||   team (https://github.com/Roman-Supernova-PIT). I extend my      ||#
#||   sincere appreciation to my team members and our Hackathon       ||#
#||   mentors for their valuable support and insightful               ||#
#||   suggestions.                                                    ||#
#||                                                                   ||#
#||   I would like to specifically acknowledge the contributions      ||#
#||   of the following team members: Lauren Aldoroty (Duke),          ||#
#||   Robert Knop (LBNL), Shu Liu (Pitt), and Michael Wood-Vasey      ||#
#||   (Pitt). Additionally, I am grateful for the guidance provided   ||#
#||   by our mentors, Marcus Manos and Lucas Erlandson from NVIDIA.   ||#
#||                                                                   ||#
#||   The collaborative environment fostered by the NASA GPU          ||#
#||   Hackathon and the Roman Supernova PIT team was essential for    ||#
#||   the development and optimization of this code.                  ||#
#||                                                                   ||#
#||   The initial cupy version of the DeCorrelation Calculator was    ||#
#||   developed by Lauren Aldoroty (Duke) and Robert Knop (LBNL) with ||#
#||   our mentors during the hackathon. The code can be found at:     ||#
#||   https://github.com/laldoroty/sfft/blob/nasa_hackathon/sfft      ||#
#||   /utils/DeCorrelationCalculator.py.                              ||#
#||                                                                   ||#
#########################################################################

class PureCupy_DeCorrelation_Calculator:
    @staticmethod
    def PCDC(NX_IMG, NY_IMG, KERNEL_GPU_JQueue, BKGSIG_JQueue, KERNEL_GPU_IQueue=[], BKGSIG_IQueue=[], 
        MATCH_KERNEL_GPU=None, REAL_OUTPUT=False, REAL_OUTPUT_SIZE=None, NORMALIZE_OUTPUT=True, VERBOSE_LEVEL=2):
        """Decorrelation Kernel Calculation in Pure Cupy"""
        NUM_I, NUM_J = len(KERNEL_GPU_IQueue), len(KERNEL_GPU_JQueue)
        assert NUM_J > 0
        if NUM_I == 0:
            MODE = 'IMAGE-STACKING'
            if NUM_J < 2:
                _error_message = 'IMAGE-STACKING MODE Requires at least 2 J-IMAGE!'
                raise Exception('MeLOn ERROR: %s' %_error_message)
            if np.sum([KERNEL_GPU is not None for KERNEL_GPU in KERNEL_GPU_JQueue]) == 0:
                _error_message = 'IMAGE-STACKING MODE Requires at least 1 non-None J-KERNEL!'
                raise Exception('MeLOn ERROR: %s' %_error_message)

        if NUM_I >= 1:
            MODE = 'IMAGE-SUBTRACTION'
            if NUM_J == 0: 
                _error_message = 'IMAGE-SUBTRACTION MODE Requires at least 1 I-IMAGE & 1 J-IMAGE!'
                raise Exception('MeLOn ERROR: %s' %_error_message)
            _Q = KERNEL_GPU_JQueue + KERNEL_GPU_IQueue + [MATCH_KERNEL_GPU]
            if np.sum([KERNEL_GPU is not None for KERNEL_GPU in _Q]) == 0:
                _error_message = 'IMAGE-SUBTRACTION MODE Requires at least 1 non-None J/I/MATCH-KERNEL!'
                raise Exception('MeLOn ERROR: %s' %_error_message)

        DELTA_KERNEL_GPU = cp.array([
            [0, 0, 0], 
            [0, 1, 0], 
            [0, 0, 0]], dtype=cp.float64
        )
        
        FDENO_GPU = None
        for KERNEL_GPU, BKGSIG in zip(KERNEL_GPU_JQueue, BKGSIG_JQueue):
            if KERNEL_GPU is not None:
                K_CSZ_GPU = PureCupy_FFTKits.KERNEL_CSZ(KERNEL_GPU=KERNEL_GPU, NX_IMG=NX_IMG, NY_IMG=NY_IMG)
            else:
                K_CSZ_GPU = PureCupy_FFTKits.KERNEL_CSZ(KERNEL_GPU=DELTA_KERNEL_GPU, NX_IMG=NX_IMG, NY_IMG=NY_IMG)
            FK_CSZ_GPU = cp.fft.fft2(K_CSZ_GPU)
            FK2_CSZ_GPU = (cp.conj(FK_CSZ_GPU) * FK_CSZ_GPU).real
            if FDENO_GPU is None:
                FDENO_GPU = (BKGSIG**2 * FK2_CSZ_GPU) / NUM_J**2
            else: 
                FDENO_GPU += (BKGSIG**2 * FK2_CSZ_GPU) / NUM_J**2
        
        if MATCH_KERNEL_GPU is not None:
            MK_CSZ_GPU = PureCupy_FFTKits.KERNEL_CSZ(KERNEL_GPU=MATCH_KERNEL_GPU, NX_IMG=NX_IMG, NY_IMG=NY_IMG)
        else:
            MK_CSZ_GPU = PureCupy_FFTKits.KERNEL_CSZ(KERNEL_GPU=DELTA_KERNEL_GPU, NX_IMG=NX_IMG, NY_IMG=NY_IMG)

        FMK_CSZ_GPU = cp.fft.fft2(MK_CSZ_GPU)
        FMK2_CSZ_GPU = (cp.conj(FMK_CSZ_GPU) * FMK_CSZ_GPU).real
        
        for KERNEL_GPU, BKGSIG in zip(KERNEL_GPU_IQueue, BKGSIG_IQueue):
            if KERNEL_GPU is not None:
                K_CSZ_GPU = PureCupy_FFTKits.KERNEL_CSZ(KERNEL_GPU=KERNEL_GPU, NX_IMG=NX_IMG, NY_IMG=NY_IMG)
            else:
                K_CSZ_GPU = PureCupy_FFTKits.KERNEL_CSZ(KERNEL_GPU=DELTA_KERNEL_GPU, NX_IMG=NX_IMG, NY_IMG=NY_IMG)
            FK_CSZ_GPU = cp.fft.fft2(K_CSZ_GPU)
            FK2_CSZ_GPU = (cp.conj(FK_CSZ_GPU) * FK_CSZ_GPU).real
            FDENO_GPU += (BKGSIG**2 * FK2_CSZ_GPU * FMK2_CSZ_GPU) / NUM_I**2

        FDENO_GPU = cp.sqrt(FDENO_GPU)
        FKDECO_GPU = 1. / FDENO_GPU

        if not REAL_OUTPUT:
            if NORMALIZE_OUTPUT:
                NORMALIZE_FACTOR = 1./FKDECO_GPU[0, 0]
                FKDECO_GPU *= NORMALIZE_FACTOR
            return FKDECO_GPU
    
        if REAL_OUTPUT:
            KDECO_GPU = cp.fft.ifft2(FKDECO_GPU).real
            KDECO_GPU = PureCupy_FFTKits.KERNEL_CSZ_INV(
                KIMG_GPU=KDECO_GPU, NX_KERN=REAL_OUTPUT_SIZE[0], NY_KERN=REAL_OUTPUT_SIZE[1],
                VERBOSE_LEVEL=VERBOSE_LEVEL
            )
            
            if NORMALIZE_OUTPUT:
                assert REAL_OUTPUT_SIZE is not None
                NORMALIZE_FACTOR = 1./cp.sum(KDECO_GPU)
                KDECO_GPU *= NORMALIZE_FACTOR
            return KDECO_GPU



import numpy as np
from sfft.utils.ReadWCS import Read_WCS

__last_update__ = "2024-09-22"
__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.6.1"

class PatternRotation_Calculator:
    @staticmethod
    def PRC(hdr_obj, hdr_targ):
        """calculate the rotation angle for a pattern from the object frame to target frame"""
        def calculate_skyN_vector(wcshdr, x_start, y_start, shift_dec=1.0):
            w = Read_WCS.RW(wcshdr, VERBOSE_LEVEL=1)
            ra_start, dec_start = w.all_pix2world(np.array([[x_start, y_start]]), 1)[0]
            ra_end, dec_end = ra_start, dec_start + shift_dec/3600.0
            x_end, y_end = w.all_world2pix(np.array([[ra_end, dec_end]]), 1)[0]
            skyN_vector = np.array([x_end - x_start, y_end - y_start])
            return skyN_vector

        # Note: rotate_angle is counterclockwise, in degree
        def calculate_rotate_angle(vector_ref, vector_obj):
            rad = np.arctan2(np.cross(vector_ref, vector_obj), np.dot(vector_ref, vector_obj))
            rotate_angle = np.rad2deg(rad)
            if rotate_angle < 0.0: rotate_angle += 360.0 
            return rotate_angle

        _phdr = hdr_obj
        _w = Read_WCS.RW(_phdr, VERBOSE_LEVEL=1)
        x0, y0 = 0.5 + int(_phdr['NAXIS1'])/2.0, 0.5 + int(_phdr['NAXIS2'])/2.0
        ra0, dec0 = _w.all_pix2world(np.array([[x0, y0]]), 1)[0]
        skyN_vector = calculate_skyN_vector(wcshdr=_phdr, x_start=x0, y_start=y0)

        _phdr = hdr_targ
        _w = Read_WCS.RW(_phdr, VERBOSE_LEVEL=1)
        x1, y1 = _w.all_world2pix(np.array([[ra0, dec0]]), 1)[0]
        skyN_vectorp = calculate_skyN_vector(wcshdr=_phdr, x_start=x1, y_start=y1)
        PATTERN_ROTATE_ANGLE = calculate_rotate_angle(vector_ref=skyN_vector, vector_obj=skyN_vectorp)

        return PATTERN_ROTATE_ANGLE



import warnings
import numpy as np
# version: Aug 17, 2022

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.3"

class Weighted_Quantile:
    @staticmethod
    def WQ(values, weights, quantiles, values_sorted=False, old_style=False):
        # ref: https://stackoverflow.com/questions/21844024/weighted-percentile-using-numpy
        """ Very close to numpy.percentile, but supports weights.
        NOTE: quantiles should be in [0, 1]!
        :param values: numpy.array with data
        :param weights: array-like of the same length as `array`
        :param quantiles: array-like with many quantiles needed
        :param values_sorted: bool, if True, then will avoid sorting of
            initial array
        :param old_style: if True, will correct output to be consistent
            with numpy.percentile.
        :return: numpy.array with computed quantiles.
        """
        values = np.array(values)
        quantiles = np.array(quantiles)
        if weights is None:
            weights = np.ones(len(values))
        weights = np.array(weights)
        assert np.all(quantiles >= 0) and np.all(quantiles <= 1), \
            'quantiles should be in [0, 1]'

        if not values_sorted:
            sorter = np.argsort(values)
            values = values[sorter]
            weights = weights[sorter]

        weighted_quantiles = np.cumsum(weights) - 0.5 * weights
        if old_style:
            # To be convenient with numpy.percentile
            weighted_quantiles -= weighted_quantiles[0]
            weighted_quantiles /= weighted_quantiles[-1]
        else:
            weighted_quantiles /= np.sum(weights)
        return np.interp(quantiles, weighted_quantiles, values)

class TopFlatten_Weighted_Quantile:
    @staticmethod
    def TFWQ(values, weights, quantiles, NUM_TOP_END=30):
        # *** Note on TopFlatten Weighted ***
        # (1) We flatten the top weights to avoid the calculation strongly biased towards 
        #     to a handful of inidividual objects with extremely high weights.
        # (2) E.g., When you use source flux as sample weight, it is possible that a few 
        #     brightest sources in the field would possess undesired dramtically high weights.
        #     In such cases, it is useful to leverage the TopFlatten to calculate weighted quantiles.

        assert len(values) > 0
        if len(values) <= NUM_TOP_END:
            # flatten all when the sample size < NUM_TOP_END
            outputs = np.percentile(values, quantiles)
            _warn_message = 'CALCULATING WEIGHTED QUANTILES --- '
            _warn_message += 'USE UNIFORM-WEIGHTED MEDIAN OVER [%d] SAMPLES!' %(len(values))
            warnings.warn('MeLOn WARNING: %s' %_warn_message)
            return outputs
        
        # flatten the top end
        topflatten = lambda W: np.clip(W/(np.sort(W)[-NUM_TOP_END]), a_min=0.0, a_max=1.0)
        outputs = Weighted_Quantile.WQ(values=values, quantiles=quantiles, weights=topflatten(weights))
        return outputs



from setuptools import setup, find_packages

DESCRIPTION = "Image Subtraction in Fourier Space"
LONG_DESCRIPTION = open('README.rst').read()
NAME = "sfft-romansnpit"
AUTHOR = "Lei Hu"
AUTHOR_EMAIL = "leihu@andrew.cmu.edu"
MAINTAINER = "Lei Hu"
MAINTAINER_EMAIL = "leihu@andrew.cmu.edu"
DOWNLOAD_URL = 'https://github.com/thomasvrussell/sfft'

LICENSE = 'MIT License'
VERSION = '1.6.4.dev9'

install_reqs = ['scipy>=1.5.2',
                'astropy>=3.2.3',
                'scikit-image>=0.25.2',
                'fastremap>=1.7.0',
                'sep>=1.0.3',
                'numba>=0.53.1',
                'llvmlite>=0.36.0',
                'pyfftw>=0.12.0']

setup(name=NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type='text/markdown',
      setup_requires=['numpy'],
      install_requires=install_reqs,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      maintainer=MAINTAINER,
      maintainer_email=MAINTAINER_EMAIL,
      download_url=DOWNLOAD_URL,
      license=LICENSE,
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering :: Astronomy'],
     )



import warnings
from astropy.wcs import WCS, FITSFixedWarning
# version: Mar 17, 2023

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.4"

class Read_WCS:
    @staticmethod
    def RW(hdr, VERBOSE_LEVEL=2):

        """
        # Remarks on the WCS in FITS header and astropy.WCS
        # Case A:
        #    FITS header: 
        #        CTYPE1 = 'RA---TAN'   CTYPE2 = 'DEC--TAN'
        #        CD1_1 = ...           CD1_2 = ...
        #        CD2_1 = ...           CD2_2 = ... 
        #        NOTE: NO DISTORTION TERMS!
        #    
        #    astropy.WCS:        
        #        # print(WCS(phr).to_header())
        #        CTYPE1 = 'RA---TAN'   CTYPE2 = 'DEC--TAN'
        #        PC1_1 = ...           PC1_2 = ...
        #        PC2_1 = ...           PC2_2 = ... 
        #        NOTE: if CD1_2 = CD2_1 = 0.0, then PC1_2 and PC2_1 will be absent.
        #
        # Case B (Distorted tangential with the TPV FITS convention, added officially to the FITS registry in Aug. 2012):
        #    FITS header: 
        #        CTYPE1 = 'RA---TPV'   CTYPE2 = 'DEC--TPV'
        #        CD1_1 = ...           CD1_2 = ...
        #        CD2_1 = ...           CD2_2 = ... 
        #        PV1_0 = ...           PV1_1 = ...
        #        ...
        #    
        #    astropy.WCS:        
        #        # print(WCS(phr).to_header())
        #        CTYPE1 = 'RA---TPV'   CTYPE2 = 'DEC--TPV'
        #        PC1_1 = ...           PC1_2 = ...
        #        PC2_1 = ...           PC2_2 = ... 
        #        PV1_0 = ...           PV1_1 = ...
        #        ...
        #        NOTE: if CD1_2 = CD2_1 = 0.0, then PC1_2 and PC2_1 will be absent.
        #
        # Case B' (Distorted tangential with Greisen & Calabretta's 2000 draft, obsoleted):
        #    FITS header:
        #        CTYPE1 = 'RA---TAN'   CTYPE2 = 'DEC--TAN'
        #        CD1_1 = ...           CD1_2 = ...
        #        CD2_1 = ...           CD2_2 = ... 
        #        PV1_0 = ...           PV1_1 = ...
        #        ...
        #    
        #    WARNING: To avoid Astropy.WCS incompatibility, one should make corrections 
        #        as follows before reading by Astropy.WCS
        #        phr['CTYPE1'] = 'RA---TPV'
        #        phr['CTYPE1'] = 'DEC--TPV'
        #        w = WCS(phr)
        #      
        #    astropy.WCS:        
        #        # print(w.to_header())
        #        CTYPE1 = 'RA---TPV'   CTYPE2 = 'DEC--TPV'
        #        PC1_1 = ...           PC1_2 = ...
        #        PC2_1 = ...           PC2_2 = ... 
        #        PV1_0 = ...           PV1_1 = ...
        #        ...
        #        NOTE: if CD1_2 = CD2_1 = 0.0, then PC1_2 and PC2_1 will be absent.
        #
        # Case C:
        #    FITS header: 
        #        CTYPE1 = 'RA---TAN-SIP'   CTYPE2 = 'DEC--TAN-SIP'
        #        CD1_1 = ...           CD1_2 = ...
        #        CD2_1 = ...           CD2_2 = ... 
        #        A_ORDER = 2           
        #        A_0_0 = ...           A_0_1 = ...
        #        ...
        #        B_ORDER = 2           
        #        B_0_0 = ...           B_0_1 = ...
        #        ...
        #        AP_ORDER = 2           
        #        AP_0_0 = ...          AP_0_1 = ...
        #        ...
        #        BP_ORDER = 2           
        #        BP_0_0 = ...          BP_0_1 = ...
        #        ...
        #    
        #    astropy.WCS:        
        #        # print(WCS(phr).to_header(relax=True))
        #        CTYPE1 = 'RA---TAN-SIP'   CTYPE2 = 'DEC--TAN-SIP'
        #        PC1_1 = ...           PC1_2 = ...
        #        PC2_1 = ...           PC2_2 = ... 
        #        A_ORDER = 2           
        #        A_0_0 = ...           A_0_1 = ...
        #        ...
        #        B_ORDER = 2           
        #        B_0_0 = ...           B_0_1 = ...
        #        ...
        #        AP_ORDER = 2           
        #        AP_0_0 = ...          AP_0_1 = ...
        #        ...
        #        BP_ORDER = 2           
        #        BP_0_0 = ...          BP_0_1 = ...
        #        ...
        #        NOTE: if CD1_2 = CD2_1 = 0.0, then PC1_2 and PC2_1 will be absent.
        #
        # P.S. One may use PC1_1 to replace CD1_1 in FITS header for above cases, not recommended though.
        #
        """

        with warnings.catch_warnings():
            if VERBOSE_LEVEL in [0, 1]: behavior = 'ignore'
            if VERBOSE_LEVEL in [2]: behavior = 'default'
            warnings.filterwarnings(behavior, category=FITSFixedWarning)

            if hdr['CTYPE1'] == 'RA---TAN' and 'PV1_0' in hdr:
                _hdr = hdr.copy()
                _hdr['CTYPE1'] = 'RA---TPV'
                _hdr['CTYPE2'] = 'DEC--TPV'
            else: _hdr = hdr
            w = WCS(_hdr)
        
        return w
